from addict import Dict
from sklearn.base import BaseEstimator
from jamboree.middleware.procedures import ModelProcedureAbstract
from sklearn.datasets import make_friedman2
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import DotProduct, WhiteKernel
from loguru import logger




class CremeProcedure(ModelProcedureAbstract):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self.requirements.model = True
        self.requirements.criterion = False
        self.requirements.optimizer = False
        
        # types = Dict()
        # types.model = BaseEstimator
        
        self.types.model = BaseEstimator
    
    @logger.catch
    def get_params(self):
        self.verify()
        return self.dictionary.model.get_params()
        
    @logger.catch
    def predict(self, X, **kwargs):
        self.verify()
        return self.dictionary.model.predict(X, **kwargs)
    
    @logger.catch
    def predict_prob(self, X, **kwargs):
        self.verify()
        return self.dictionary.model.predict_prob(X, **kwargs)

    @logger.catch
    def partial_fit(self, X, y, **kwargs):
        self.verify()
        self.dictionary.model.partial_fit(X, y, **kwargs)

    def fit(self, X, y, **kwargs):
        self.verify()
        self.dictionary.model.fit(X, y, **kwargs)
        # print(self.mdict.model.predict(X[:2,:], return_std=True))



def main():
    import datetime as dt
    from creme import compose
    from creme import datasets
    from creme import feature_extraction
    from creme import linear_model
    from creme import metrics as metricss
    from creme import preprocessing
    from creme import stats
    from creme import stream


    X_y = datasets.Bikes()
    X_y = stream.simulate_qa(X_y, moment='moment', delay=dt.timedelta(minutes=30))

    def add_time_features(x):
        return {
            **x,
            'hour': x['moment'].hour,
            'day': x['moment'].weekday()
        }

    model = add_time_features
    model |= (
        compose.Select('clouds', 'humidity', 'pressure', 'temperature', 'wind') +
        feature_extraction.TargetAgg(by=['station', 'hour'], how=stats.Mean()) +
        feature_extraction.TargetAgg(by='station', how=stats.EWMean())
    )
    model |= preprocessing.StandardScaler()
    model |= linear_model.LinearRegression()

    metric = metricss.MAE()

    questions = {}

    for i, x, y in X_y:
        # Question
        is_question = y is None
        if is_question:
            y_pred = model.predict_one(x)
            questions[i] = y_pred

        # Answer
        else:
            metric.update(y, questions[i])
            model = model.fit_one(x, y)

            if i >= 30000 and i % 30000 == 0:
                print(i, metric)

if __name__ == "__main__":
    main()


from addict import Dict
from sklearn.base import BaseEstimator
from jamboree.middleware.procedures import ModelProcedureAbstract
from sklearn.datasets import make_friedman2
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import DotProduct, WhiteKernel
from loguru import logger


""" TODO: FIX THIS CRAP!!!"""


class TFKerasProcedure(ModelProcedureAbstract):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self.requirements.model = True
        self.requirements.criterion = False
        self.requirements.optimizer = False
        
        # types = Dict()
        # types.model = BaseEstimator
        
        self.types.model = BaseEstimator
    
    @logger.catch
    def get_params(self):
        self.verify()
        return self.dictionary.model.get_params()
        
    @logger.catch
    def predict(self, X, **kwargs):
        self.verify()
        return self.dictionary.model.predict(X, **kwargs)
    
    @logger.catch
    def predict_prob(self, X, **kwargs):
        self.verify()
        return self.dictionary.model.predict_prob(X, **kwargs)

    @logger.catch
    def partial_fit(self, X, y, **kwargs):
        self.verify()
        self.dictionary.model.partial_fit(X, y, **kwargs)

    def fit(self, X, y, **kwargs):
        self.verify()
        self.dictionary.model.fit(X, y, **kwargs)
        # print(self.mdict.model.predict(X[:2,:], return_std=True))


import datetime
from pathlib import Path
from typing import Callable, Optional, Union

import numpy as np
import pandas as pd
import pytest
from loguru import logger
from pandas.api.types import is_datetime64_any_dtype as is_datetime

from darwin_ml import __version__
from darwin_ml.utils import Windowing
from darwin_ml.utils.preprocessing import (format_look_ahead,
                                           format_timeseries_dataframe)


def roll_dataframe_stats(frame: pd.DataFrame,
                         window=14,
                         min_steps: int = 1,
                         callback: Optional[Callable] = None):
    windower = Windowing(frame, window_size=window)

    step_count = 0
    history = []

    while windower.has_next_observation:
        res = windower.step()
        if callback is not None:
            history.append(callback(res))
        step_count += 1

    return step_count >= min_steps, history



@pytest.fixture
def monthly_sunspot_data():
    BASE_DIR = Path(__file__).resolve().parent.parent.cwd(
    ) / 'data' / 'monthly-sunspots.csv'
    BASE_DIR_STR = str(BASE_DIR)
    return pd.read_csv(BASE_DIR_STR)


@pytest.fixture
def stock_data():
    BASE_DIR = Path(
        __file__).resolve().parent.parent.cwd() / 'data' / 'stock_data.csv'
    BASE_DIR_STR = str(BASE_DIR)
    return pd.read_csv(BASE_DIR_STR)



def test_check_sunspot_dataframe(monthly_sunspot_data):
    logger.debug(monthly_sunspot_data)
    assert isinstance(monthly_sunspot_data, pd.DataFrame)


def test_check_stock_data(stock_data):
    columns = list(stock_data.columns)
    stock_data_len = len(stock_data)

    # Logs just in case the data fails
    logger.debug(stock_data)
    logger.debug(columns)

    assert isinstance(stock_data, pd.DataFrame)
    assert columns == [
        'Timestamp', 'Open', 'High', 'Low', 'Close', 'Volume_BTC',
        'Volume_Currency', 'Weighted_Price'
    ], "Stock Data Is Found"
    assert stock_data_len > 0, "The stock dataframe is empty"

def test_timeseries_indexing(stock_data):
    columns = list(stock_data.columns)
    with pytest.raises(AttributeError):
        format_timeseries_dataframe(stock_data, "DERP")
    
    df = format_timeseries_dataframe(stock_data, "Timestamp")
    # logger.info(df)
    index_dtype = df.index.dtype
    assert len(columns) != len(df.columns)
    assert is_datetime(index_dtype)

def test_rolling_window_frame(stock_data):
    is_min, history = roll_dataframe_stats(stock_data)

    assert len(
        history
    ) == 0, "There shouldn't have been a callback function added. History should be empty"
    assert is_min == True, "Looks like the dataframe wasn't stepped through at least one time"

# def test_rolling_window_frame(stock_data):
#     is_min, history = roll_dataframe_stats(stock_data)

#     assert len(
#         history
#     ) == 0, "There shouldn't have been a callback function added. History should be empty"
#     assert is_min == True, "Looks like the dataframe wasn't stepped through at least one time"


def test_regression_model(stock_data):
    assert True, "Model has not been set"
    assert len([1, 2, 3, 4]) > 0, "Metrics weren't added"
    assert len([1, 2, 3, 4]) > 0, "Metrics weren't added"


import numpy as np
import pandas as pd
from pydantic import BaseModel
from creme import (anomaly, cluster, ensemble, linear_model, metrics,
                   naive_bayes, optim, preprocessing, proba, stats)
from loguru import logger

class AllocationModel(BaseModel):
    # Get the average percentage change of the allocation
    rolling_mean_3: float = 0.0
    rolling_mean_5: float = 0.0
    rolling_mean_7: float = 0.0
    rolling_mean_10: float = 0.0
    rolling_sem_3: float = 0.0
    rolling_sem_5: float = 0.0
    rolling_sem_7: float = 0.0
    rolling_sem_10: float = 0.0
    rolling_var_3: float = 0.0
    rolling_var_5: float = 0.0
    rolling_var_7: float = 0.0
    rolling_var_10: float = 0.0
    auto_regression_2: float = 0.0
    auto_regression_4: float = 0.0
    auto_regression_6: float = 0.0
    pct_filled_2: float = 0.0
    pct_filled_4: float = 0.0
    pct_filled_6: float = 0.0
    auto_corr_pct_filled_2: float = 0.0
    auto_corr_pct_filled_4: float = 0.0
    auto_corr_pct_filled_6: float = 0.0

    class Config:
        arbitrary_types_allowed = True

    def space(self):
        arr = [
            self.rolling_mean_3,
            self.rolling_mean_5,
            self.rolling_mean_7,
            self.rolling_mean_10,
            self.rolling_sem_3,
            self.rolling_sem_5,
            self.rolling_sem_7,
            self.rolling_sem_10,
            self.rolling_var_3,
            self.rolling_var_5,
            self.rolling_var_7,
            self.rolling_var_10,
            self.auto_regression_2,
            self.auto_regression_4,
            self.auto_regression_6,
            self.pct_filled_2,
            self.pct_filled_4,
            self.pct_filled_6,
        ]
        return np.array(arr)


class AllocationStatsVariates(BaseModel):
    rolling_mean_3 = stats.RollingMean(window_size=3)
    rolling_mean_5 = stats.RollingMean(window_size=5)
    rolling_mean_7 = stats.RollingMean(window_size=7)
    rolling_mean_10 = stats.RollingMean(window_size=10)

    rolling_sem_3 = stats.RollingSEM(window_size=3)
    rolling_sem_5 = stats.RollingSEM(window_size=5)
    rolling_sem_7 = stats.RollingSEM(window_size=7)
    rolling_sem_10 = stats.RollingSEM(window_size=10)

    rolling_var_3 = stats.RollingVar(window_size=3)
    rolling_var_5 = stats.RollingVar(window_size=5)
    rolling_var_7 = stats.RollingVar(window_size=7)
    rolling_var_10 = stats.RollingVar(window_size=10)

    auto_regression_3 = stats.AutoCorrelation(lag=3)
    auto_regression_4 = stats.AutoCorrelation(lag=4)
    auto_regression_5 = stats.AutoCorrelation(lag=5)

    mean_pct_filled_2 = stats.RollingMean(window_size=2)
    mean_pct_filled_4 = stats.RollingMean(window_size=4)
    mean_pct_filled_6 = stats.RollingMean(window_size=6)

    auto_corr_pct_filled_2 = stats.AutoCorrelation(lag=2)
    auto_corr_pct_filled_4 = stats.AutoCorrelation(lag=4)
    auto_corr_pct_filled_6 = stats.AutoCorrelation(lag=6)

    class Config:
        arbitrary_types_allowed = True

    def zero_complex(self, num):
        if isinstance(num, complex): return 0.0
        return num

    def get_model(self) -> AllocationModel:
        return AllocationModel(
              rolling_mean_3         = self.rolling_mean_3.get(),
              rolling_mean_5         = self.rolling_mean_5.get(),
              rolling_mean_7         = self.rolling_mean_7.get(),
              rolling_mean_10        = self.rolling_mean_10.get(),
              rolling_sem_3          = self.zero_complex(self.rolling_sem_3.get()),
              rolling_sem_5          = self.zero_complex(self.rolling_sem_5.get()),
              rolling_sem_7          = self.zero_complex(self.rolling_sem_7.get()),
              rolling_sem_10         = self.rolling_sem_10.get(),
              rolling_var_3          = self.rolling_var_3.get(),
              rolling_var_5          = self.rolling_var_5.get(),
              rolling_var_7          = self.rolling_var_7.get(),
              rolling_var_10         = self.rolling_var_10.get(),
              auto_regression_2      = self.auto_regression_3.get(),
              auto_regression_4      = self.auto_regression_4.get(),
              auto_regression_6      = self.auto_regression_5.get(),
              pct_filled_2           = self.mean_pct_filled_2.get(),
              pct_filled_4           = self.mean_pct_filled_4.get(),
              pct_filled_6           = self.mean_pct_filled_6.get(),
              auto_corr_pct_filled_2 = self.auto_corr_pct_filled_2.get(),
              auto_corr_pct_filled_4 = self.auto_corr_pct_filled_4.get(),
            auto_corr_pct_filled_6   = self.zero_complex(self.auto_corr_pct_filled_6.get())
        )


class AllocationStepSystem(object):
    def __init__(self):
        self.allo_variates: AllocationStatsVariates = AllocationStatsVariates()

    def update_pct(self, _pct: float):
        self.allo_variates.rolling_mean_3.update(_pct)
        self.allo_variates.rolling_mean_5.update(_pct)
        self.allo_variates.rolling_mean_7.update(_pct)
        self.allo_variates.rolling_mean_10.update(_pct)
        self.allo_variates.rolling_sem_3.update(_pct)
        self.allo_variates.rolling_sem_5.update(_pct)
        self.allo_variates.rolling_sem_7.update(_pct)
        self.allo_variates.rolling_sem_10.update(_pct)
        self.allo_variates.rolling_var_3.update(_pct)
        self.allo_variates.rolling_var_5.update(_pct)
        self.allo_variates.rolling_var_7.update(_pct)
        self.allo_variates.rolling_var_10.update(_pct)
        self.allo_variates.auto_regression_3.update(_pct)
        self.allo_variates.auto_regression_4.update(_pct)
        self.allo_variates.auto_regression_5.update(_pct)

    def update_fill(self, _fill_pct: float):
        self.allo_variates.mean_pct_filled_2.update(_fill_pct)
        self.allo_variates.mean_pct_filled_4.update(_fill_pct)
        self.allo_variates.mean_pct_filled_6.update(_fill_pct)
        self.allo_variates.auto_corr_pct_filled_2.update(_fill_pct)
        self.allo_variates.auto_corr_pct_filled_4.update(_fill_pct)
        
        self.allo_variates.auto_corr_pct_filled_6.update(_fill_pct)

    def step(self, allocation_pct: float,
             allocation_fill: float) -> AllocationModel:
        
        self.update_fill(allocation_fill)
        self.update_pct(allocation_pct)
        return self.allo_variates.get_model()


import os
from jamboree import Jamboree
REDIS_HOST = os.getenv("REDIS_HOST", "0.0.0.0")
REDIS_PORT = int(os.getenv("REDIS_PORT", "3979"))


def setup_jamboree():
    return Jamboree(REDIS_HOST=REDIS_HOST, REDIS_PORT=REDIS_PORT)


import numpy as np
import pandas as pd
from creme import stats
from pydantic import BaseModel
import collections
from darwin_ml.utils import Injectable
from darwin_ml.models.creme_ml.transforms import PercentageChange
from darwin_ml.models.creme_ml.transforms import FeatureWindow
from darwin_ml.middleware import containers


class PriceResponse(BaseModel):
    """A return model for the price delta

    Args:
        BaseModel (None): This is the model we're getting returned.
    """
    direction: int = 0
    percentage: float = 0.0


class PriceDeltaModel(BaseModel, Injectable):
    # Get the average percentage change of the allocation
    pct_model: PercentageChange = PercentageChange(lag=3)

    class Config:
        arbitrary_types_allowed = True

    def set_lag(self, lag=3):
        self.pct_model = PercentageChange(lag=lag)

    def stored_model(self, price_container: containers.PriceContainer):
        price: float = price_container.current_price
        self.pct_model.update(price)

        pct_window = self.pct_model.get()
        direction = self.pct_model.get_class()

        res = PriceResponse(direction=direction, percentage=pct_window)
        return res

    def step(self,
             price_container: containers.PriceContainer,
             is_stored: bool = False) -> PriceResponse:
        """Add price information to get target change information.

        Args:
            price_dict (dict): OHLCV Price Information.

        Returns:
            PriceResponse: The direction classification of the price change over a gievn lag.
        """
        
        if is_stored:
            return self.stored_model(price_container)

        direction = price_container.direction
        percentage = price_container.percentage

        res = PriceResponse(direction=direction, percentage=percentage)
        return res


from addict import Dict
from loguru import logger
from typing import Optional

from sklearn.base import BaseEstimator
from sklearn.datasets import make_friedman2
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import DotProduct, WhiteKernel
from jamboree.middleware.procedures import ModelProcedureAbstract

from skorch.net import NeuralNet

from torch.nn import Module
from torch.nn.modules.loss import _Loss
from torch.optim import Optimizer
from torch.optim import Adam
import numpy as np
from sklearn.datasets import make_classification
from torch import nn
import torch.nn.functional as F

from skorch import NeuralNetClassifier


class TorchProcedure(ModelProcedureAbstract):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self.requirements.model = True
        self.requirements.criterion = True
        self.requirements.optimizer = True
        self.types.model = nn.Module
        self.types.criterion = _Loss
        self.types.optimizer = Optimizer
        self._compiled_model:Optional[NeuralNet] = None
    
    @property
    def model(self) -> NeuralNet:
        self.verify()
        if self._compiled_model is None:
            _compiled = NeuralNet(
                module=self.dictionary.model,
                criterion=self.dictionary.criterion,
                optimizer=self.dictionary.optimizer,
                max_epochs=10,
                lr=0.1,
                # Shuffle training data on each epoch
                iterator_train__shuffle=True,
            )
            self._compiled_model = _compiled
        return self._compiled_model

    @logger.catch
    def set_params(self, **params):
        self.changed = True
        self.model.set_params(**params)

    @logger.catch
    def get_params(self):
        return self.model.get_params()
        
    @logger.catch
    def predict(self, X, **kwargs):
        return self.model.predict(X, **kwargs)
    
    @logger.catch
    def predict_proba(self, X, **kwargs):
        return self.model.predict_proba(X, **kwargs)

    @logger.catch
    def partial_fit(self, X, y, **kwargs):
        self.changed = True
        self.model.partial_fit(X, y, **kwargs)

    def fit(self, X, y, **kwargs):
        self.changed = True
        self.model.fit(X, y, **kwargs)


class MyModule(nn.Module):
    def __init__(self, num_units=10, nonlin=F.relu):
        super(MyModule, self).__init__()

        self.dense0 = nn.Linear(20, num_units)
        self.nonlin = nonlin
        self.dropout = nn.Dropout(0.5)
        self.dense1 = nn.Linear(num_units, 10)
        self.output = nn.Linear(10, 2)

    def forward(self, X, **kwargs):
        X = self.nonlin(self.dense0(X))
        X = self.dropout(X)
        X = F.relu(self.dense1(X))
        X = F.softmax(self.output(X))
        return X


class TestCustomTorchClassifier(TorchProcedure):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dictionary.model = MyModule
        self.dictionary.optimizer = Adam
        self.dictionary.criterion = nn.NLLLoss




def main():
    


    X, y = make_classification(1000, 20, n_informative=10, random_state=0)
    X = X.astype(np.float32)
    y = y.astype(np.int64)


    net = TestCustomTorchClassifier()

    net.fit(X, y)
    for _ in range(10):
        y_proba = net.predict_proba(X)
        print(y_proba)
    

if __name__ == "__main__":
    main()


import numpy as np
from typing import Optional, Dict, Any
from jamboree.utils import omit

class FeaturesContainer(object):
    def __init__(self, latest_feature: Dict[str, Any], prior_feature: Optional[Dict[str, Any]] = None):

        self._current = self.remove_timestamps(latest_feature)
        self._prior:Optional[Dict[str, Any]] = None
        
        
        if bool(prior_feature):
            self._prior = self.remove_timestamps(prior_feature, name="PriorFeature")
    
    def is_dict(self, _feature:dict, name:str="CurrentFeature"):
        assert isinstance(_feature, dict), f"{name} is not a dict"

    @property
    def current(self) -> Dict[str, Any]:
        """Get the latest bar.

        Returns:
            Dict[str, Any]: The latest bar.
        """
        # self.verify()
        return self._current

    @property
    def previous(self) -> Dict[str, Any]:
        """The previous price bar.

        Returns:
            Dict[str, Any]: The previous price bar.
        """
        # self.verify(is_prev=True)
        return self._prior

    @property
    def is_none(self) -> bool:
        """Checks if both variables are empty.

        Returns:
            bool: Are both variables there.
        """
        return (not bool(self.current)) and (not bool(self.previous))

    @property
    def is_current(self) -> bool:
        """Determines if the event has only the current bar.

        Returns:
            bool: if this is the current bar.
        """
        return bool(self.current)

    @property
    def is_both(self) -> bool:
        """Determines if both features are there

        Returns:
            bool: if this is the current bar.
        """
        return bool(self.current) and bool(self.previous)
    @property
    def is_previous(self) -> bool:
        return bool(self.previous)
    
    @property
    def features(self):
        curr = []
        if self.is_current:
            curr = list(self.current.values())
        return np.array(curr)


    def verify(self, is_both: bool = False, is_prev: bool = False) -> None:
        """Run through a series of verifications.

        1. Check if at least one of the variables have been set.
        2. If we're checking for both, yet both variables aren't there, we should exit.
        3. We're checking if the latest price information is there.

        

        Args:
            is_both (bool, optional): Determines if we're checking for both variables. Defaults to False.
            is_prev (bool, optional): Determines if we're checking for the previous bar. Defaults to False.

        Raises:
            AttributeError: None of the variables have been set.
            AttributeError: One of the two expected values isn't there.
            AttributeError: The current price isn't there.
        """
        # If none of the variables have been set. We raise an attribute error.
        if not self.is_none:
            raise AttributeError("None of the variables have been set.")

        # If we're looking for both variables, and one isn't set, we raise a value error
        if (is_both or is_prev) and not self.is_both:
            raise AttributeError("One of the two expected values isn't there.")

        # Check if the current bar is here
        if not self.is_current:
            raise AttributeError("The current price isn't there.")
    
    def remove_timestamps(self, feature_dict:Dict[str, Any], name="CurrentFeature"):
        self.is_dict(feature_dict, name=name)
        return omit(['timestamp', 'time'], feature_dict)
    
    def __str__(self):
        features = self.features
        feats = []
        if len(features) > 3:
            feats = features[-3:]
        return f"<PriceContainer is_previous={self.is_previous} is_current={self.is_current} features={feats} >"


import math
import random
import typing
from datetime import datetime
from typing import Optional, Union
from uuid import UUID, uuid4

import numpy as np
import pandas as pd
from creme import (anomaly, cluster, compose, datasets, ensemble,
                   feature_extraction, feature_selection, linear_model, meta,
                   metrics, model_selection, naive_bayes, optim, preprocessing,
                   proba, sampling, stats)
from creme.meta import BoxCoxRegressor
from loguru import logger

from darwin_ml.models.creme_ml.transforms import PercentageChange

# logger.disable("jamboree")


class PctHardRegressorModel(object):
    def __init__(self):
        lin_reg = linear_model.LinearRegression()
        # boxcot_reg = meta.TransformedTargetRegressor(regressor=lin_reg, func=math.log, inverse_func=math.exp)
        self.model = compose.Pipeline(
            ('scale', preprocessing.StandardScaler()),
            ('lin_reg',
             sampling.HardSamplingRegressor(
                 size=400,
                 p=0.2,
                 regressor=lin_reg
                )
            )
        )

    def fit_one(self, x: dict, y: typing.Union[float, int]):
        self.model.fit_one(x, y)

    def predict_one(self, x: dict):
        return self.model.predict_one(x)



from typing import Optional, Dict, Any
from jamboree.utils import omit

class StepSettings(object):
    def __init__(self, user_id:str, episode:str, is_live=False):
        self.__user_id:str = user_id
        self.__episode:str = episode
        self.__live:bool = is_live
        self.__asset: Optional[str] = None
        

    @property
    def user_id(self) -> str:
        """The user_id of the asset we're looking for.

        Returns:
            str: The user_id
        """
        return self.__user_id
    
    @property
    def episode(self) -> str:
        return self.__episode
    

    @property
    def live(self) -> bool:
        return self.__live
    

    @property
    def asset_id(self) -> str:
        return self.__asset

    @asset_id.setter
    def asset_id(self, _asset:str):
        self.__asset = _asset


from typing import List

from addict import Dict as ADDICT
from jamboree.middleware.procedures import (ModelProcedureAbstract,
                                            ProcedureManagement)
from loguru import logger

from darwin_ml.middleware.procedures.models import (CremeProcedure,
                                                    SklearnProcedure)


class ModelProcedureManagement(ProcedureManagement):
    def __init__(self):
        super().__init__()
        self.required_attributes = ["model_type", "online"]
        
        # Don't forget to look at the 

    @property
    def allowed(self) -> List[str]:
        return ["sklearn", "creme"]

    
    def access(self, key):
        self.check_allowed(key)
        spec = {
            "sklearn": SklearnProcedure(),
            "creme": CremeProcedure(),
        }.get(key)
        return spec


if __name__ == "__main__":
    modelmgt = ModelProcedureManagement()
    print(modelmgt.access("sklearn"))



import numpy as np
import ta
import pandas as pd


def all_ta(df:pd.DataFrame, target="Close", window=14):
    """Set all of the relavent TA indicators here. This is an aggregate function.

    Arguments:
        df {pd.DataFrame} -- [description]
    """
    return df


from typing import Optional, Dict, Any
from jamboree.utils import omit

class AllocationContainer(object):
    def __init__(self, pct: float, fill: float):
        assert isinstance(pct, float), "Percentage is not a float."
        assert isinstance(fill, float), "The percentage filled is not a float. Please provide one."
        self._percentage:float = pct
        self._filling:float = fill
    


    @property
    def percentage(self) -> float:
        """Get the latest bar.

        Returns:
            float: The percentage cap of the 
        """
        return self._percentage

    @property
    def filling(self) -> float:
        """The percentage of the allocation filled.

        Returns:
            float: The percentage of the allocation filled. Should be greater than 0, yet less than 1.2. Symbolizing 120%.
        """
        return self._filling

    def __str__(self):
        return f"<AllocationContainer percentage={self.percentage} filling={self.filling}>"
    


import time
import typing
import uuid

import matplotlib.pyplot as plt
import maya
import pandas as pd
import pandas_datareader.data as web
import pytest
from jamboree import Jamboree
from jamboree.handlers import DataHandler
from loguru import logger
from stochastic.continuous import FractionalBrownianMotion

from darwin_ml.handlers import TAPredictionTransformer
from darwin_ml.params import PriceSetupResponse, DefaultPriceMetadataParameters, setup_jamboree

default_params = DefaultPriceMetadataParameters()


@anycache(cachedir='/tmp/anycache.my')
def get_dataset_info(ds_str:str):
    ticker = yf.Ticker(ds_str)
    
    return ticker.info

@anycache(cachedir='/tmp/anyticker.my')
def get_ticker_data(ds_str:str):
    data = yf.download(ds_str, group_by = 'ticker').dropna()
    return data

# Find all datasets by market
class TestSearchTrial(object):
    """ 
        # TEST SEARCH TRIAL

        Class to test running through various scenarios to:

        1. Download data (apple and msft data)
        2. Save it to given metadata inside of redis
        3. Find all various metadata inside of redis using incomplete information
        4. Get the dataset_ids by subtype match
        5. Get datasets by_ids (resemble putting datasets by)
        6. Add datasets into a multidata container
        7. Backtest through the multiple datasets
    """
    def __init__(self):
        # List of datasets we intend to download from yahoo finance
        self.datalist = ["MSFT", "AAPL", "GOOG"]
        self.datasets = []
        self.dataset_info = {}
        self.processor = Jamboree()
        self.pricedata_management = PriceData()
        self.pricedata_management.processor = self.processor
        self.dow_metadata = []
        self.reloaded_datasets = []
        self.episodes = []
        self._raw = []
        self.set_name = "default_test"
    
        

        # # pass
    
    def download_yfinance_info(self):
        for ds in self.datalist:
            info = get_dataset_info(ds)
            self.dataset_info[ds] = {
                "long_name": info["longName"],
            }


    def download_from_yfinance(self):
        self.ds_str = " ".join(self.datalist)
        logger.info(f"starting download for datasets from yahoo finance {self.ds_str}")
        self.datasets = get_ticker_data(self.ds_str)
        logger.success(f"finished downloading from yahoo finance: {self.ds_str}")
    
    def save_datasets(self):
        """ Save all of the datasets with the given metadata information. """

        for k, v in self.dataset_info.items():
            abbv = k
            long_name = v['long_name']
            df = self.datasets[k]
            self.pricedata_management.build(name=long_name, abbv=abbv, exchange="Dow Jones")
            self.pricedata_management.reset()
            self.pricedata_management.store_time_df(df, is_bar=True)
    
    def find_all_datasets_by_exchange(self):
        name = "do*"
        logger.info(f'Finding metadata by partial exchange query \"{name}\"' )
        self.dow_metadata = self.pricedata_management.by_exchange(name)
        logger.success(f"Successfully loaded metadata ... ")
    
    def find_meta_by_id(self, _id):
        """ Find a single metadata document by id"""
        return self.pricedata_management.search.pick(_id)
    
    def from_meta(self, item:dict):
        """ Gets a dataset from metadata """
        dd = Dict(**item)
        entity = dd.metatype
        name = dd.name
        category = dd.category
        subcategories = dd.subcategories
        submetatype = dd.submetatype # the specific kind of metadata (price, economic)
        abbv = dd.abbreviation

        dataset = DataHandler()
        # dataset.entity = entity
        dataset['name'] = name
        dataset['category'] = category
        dataset['subcategories'] = subcategories
        dataset['submetatype'] = submetatype
        dataset['abbreviation'] = abbv
        dataset.processor = self.processor
        dataset.reset()
        
        return dataset
    
    def raw_dict(self, item:dict):
        dd = Dict(**item)
        norm = Dict()
        name = dd.name
        category = dd.category
        subcategories = dd.subcategories
        submetatype = dd.submetatype # the specific kind of metadata (price, economic)
        abbv = dd.abbreviation
        norm.name = name
        norm.category = category
        norm.subcategories = subcategories
        norm.submetatype = submetatype
        norm.abbreviation = abbv
        return norm.to_dict()

    def get_datasets_by_metadata(self):
        """ Get all of dataset objects by metdata. Used to see if we can """
        data_list:List[DataHandler] = []
        dow_len = len(self.dow_metadata)
        
        if dow_len > 0:
            logger.debug(f"Number of metadata sets: {dow_len}")
            for _ in self.dow_metadata:
                dataset = self.from_meta(_)
                data_list.append(dataset)
                self._raw.append(self.raw_dict(_))

        ll = len(data_list)    
        self.reloaded_datasets = data_list
        logger.success(f"Successfully created datasets ... count: {ll}")
    
    def create_episodes(self):
        """ Create a bunch of episodes """
        self.episodes = [uuid.uuid4().hex for x in range(len(self.reloaded_datasets))]
        for index, episode in enumerate(self.episodes):
            dataset = self.reloaded_datasets[index]
            dataset.episode = episode
            dataset.live = False
            dataset.time.head = maya.now().subtract(weeks=600)._epoch
            dataset.time.change_stepsize(microseconds=0, days=1, hours=0)
            dataset.time.change_lookback(microseconds=0, weeks=4, hours=0)
            self.reloaded_datasets[index] = dataset
    
    def create_multi_data(self):
        self.multi_data = MultiDataManagement()
        self.multi_data["name"] = self.set_name
        self.multi_data["category"] = "Sample Category"
        self.multi_data["subcategories"] = {"extra": "info"}
        self.multi_data["submetatype"] = "price"
        self.multi_data["abbreviation"] = "PRC"
        self.multi_data.processor = self.processor
        self.multi_data.episode = uuid.uuid4().hex
        self.multi_data.reset()
        self.multi_data.add_multiple_data_sources(self._raw)
        pprint(self.multi_data.sources)


    def run_random_dataset(self):
        dataset = random.choice(self.reloaded_datasets)
        dataset.is_robust_closest = True
        self.state
        start_time = maya.now()._epoch
        i = 0
        while dataset.is_next:
            # This cuts loop time in half immediately
            dataset.time.step()
            data = dataset.closest_head(is_robust=True)
            logger.info(data)
            time_diff = maya.now().epoch - start_time
            i += 1
            milli = (1000 / (i / time_diff)) 
            logger.info(f"Number of milliseconds: {milli}")
    
    def run_multidata(self):
        self.multi_data.time.head = maya.now().subtract(weeks=600, hours=14)._epoch
        self.multi_data.time.change_stepsize(microseconds=0, days=1, hours=0)
        self.multi_data.time.change_lookback(microseconds=0, weeks=4, hours=0)
        self.multi_data.sync()
        start_time = maya.now()._epoch
        i = 0
        while self.multi_data.is_next:
            latest = self.multi_data.step("current")
            self.multi_data.time.step()
            time_diff = maya.now().epoch - start_time
            i += 1
            milli = (1000 / (i / time_diff)) 
            logger.debug(latest)
            logger.info(f"Number of milliseconds: {milli}")
            

    def run(self, is_download=True):

        if is_download:
            # Downloads the data from yahoo finance to our system
            self.download_yfinance_info()
            self.download_from_yfinance()
            self.save_datasets()
        
        # Finds every Dow Jones exchanges and stores metadata locally
        self.find_all_datasets_by_exchange() 
        # Takes local metadata and finds the datasources in the system
        self.get_datasets_by_metadata() 
        # Creates a bunch of backtest episodes (only to iterate through episodes)
        self.create_episodes()
        # Creates a batch dataset (pulling everything at once)
        self.create_multi_data()
        logger.success(self.multi_data.metaid)
        # Picks a random individual dataset and runs through it
        self.run_random_dataset()

    def run_findselection(self):
        self.find_all_datasets_by_exchange()
        item = random.choice(self.dow_metadata)
        item = Dict(**item)
        refind = self.find_meta_by_id((item.id))
        if refind is None:
            raise AttributeError("item doesn't exist")
        datasource = self.from_meta(refind)
        num = datasource.count()
        last = datasource.closest_head()
        # many = datasource.many()
        logger.warning(num)
        logger.warning(last)
        # logger.warning(many)
    

def save_data(handler:DataHandler, name:str, abbv:str, frame:pd.DataFrame):
    handler['name'] = name
    handler['abbv'] = abbv
    handler.reset()
    handler.store_time_df(frame, is_bar=True)

def download_test_data():
    microsoft = web.DataReader(
        "MSFT", "yahoo", start="2010/1/1", end="2020/5/30"
    ).round(4)

    
    apple = web.DataReader(
        "AAPL", "yahoo", start="2010/1/1", end="2020/5/30"
    ).round(4)
    return microsoft, apple

@pytest.fixture
def connect():
    return setup_jamboree()

@pytest.fixture
def phandler() -> DataHandler:
    episode_id = uuid.uuid4().hex
    data_hander = DataHandler()
    data_hander.episode = episode_id
    data_hander.live = False
    data_hander["category"] = default_params.category
    data_hander["subcategories"] = default_params.subcategories
    data_hander["submetatype"] = default_params.submetatype
    return data_hander


@pytest.fixture
def setup_full_price(connect, phandler) -> PriceSetupResponse:
    phandler.processor = connect
    microsoft, apple = download_test_data()
    response = PriceSetupResponse()
    response.microsoft = microsoft
    response.apple = apple


    return response



"""
    # State Space Modelling and Testing

    The goal of this file is to test how iterating through the state space will work.

"""
import random
import time
import uuid
from pprint import pprint
from random import uniform
from typing import Optional

import maya
from creme import datasets, stream
from jamboree import Jamboree
from jamboree.handlers import TimeHandler
from loguru import logger
from see137.handlers.exchange import PortfolioHandler
from stochastic.continuous import FractionalBrownianMotion

from darwin_ml.handlers import TAPredictionTransformer, TAPriceData
from darwin_ml.middleware.containers import (AllocationContainer,
                                             FeaturesContainer, PriceContainer)
from darwin_ml.models import OnlineModels, PriceDeltaModel, StateSpaceModel
from darwin_ml.utils import Injectable, InjectableTime


def create_samples(n, mult=10, _add=100):
    fbm = FractionalBrownianMotion(hurst=0.7, t=1)
    s = fbm.sample(n)
    s = (s * mult) + _add
    return s


def create_allocation_samples(n):
    fbm = FractionalBrownianMotion(hurst=0.7, t=1)
    allocation = fbm.sample(n)
    pct_fill = [uniform(0, 1) for _ in range(n + 1)]

    return allocation, pct_fill


def single_feature():
    return {
        "feature_1": random.uniform(0, 1),
        "feature_2": random.uniform(0, 1),
        "feature_3": random.uniform(0, 1),
        "feature_4": random.uniform(0, 1),
    }


def create_fake_features(n):
    feature_set = [single_feature() for _ in range(n + 1)]
    return feature_set


def setup_time():
    new_time = maya.now().subtract(years=5)
    return new_time._epoch


class MainStateSpace(InjectableTime):
    def __init__(self):
        super().__init__()
        # Need to create a save and load script for this.
        self.__space: StateSpaceModel = StateSpaceModel()
        self.__technical: TAPredictionTransformer = TAPredictionTransformer()
        self.__time: TimeHandler = TimeHandler()
        self.__portfolio = PortfolioHandler()

        self.exchange: str = "binance"
        self.user_id: str = uuid.uuid4().hex
        self.is_mock: bool = False

    @property
    def is_next(self) -> bool:
        is_features = self.features.is_next
        is_price = self.price.is_next

        return is_features and is_price

    @property
    def space(self) -> StateSpaceModel:
        """ Should probably manage"""
        return self.__space

    @property
    def features(self) -> TAPriceData:
        return self.technical.features

    @property
    def price(self) -> TAPriceData:
        return self.technical.price

    @property
    def portfolio(self):
        self.__portfolio.processor = self.processor
        self.__portfolio.exchange = self.exchange
        self.__portfolio.user_id = self.user_id
        self.__portfolio.episode = self.episode
        self.__portfolio['live'] = self.live
        self.__portfolio.reset()
        return self.__portfolio

    @property
    def technical(self) -> TAPredictionTransformer:
        self.__technical.processor = self.processor
        self.__technical.asset = self.asset
        self.__technical.episode = self.episode
        self.__technical.live = self.live
        return self.__technical

    @property
    def current_time(self) -> float:
        return self.time.head

    @property
    def feature_count(self) -> int:
        return self.features.count()

    @property
    def price_count(self) -> int:
        return self.price.count()

    @property
    def price_container(self) -> PriceContainer:
        current_price_set = (
            self.price
                .closest_head_omitted()
        )
        current_price_set_back = (
            self.price
                .closest_peakback_by_omitted()
        )

        price_container = PriceContainer(
            current_price_set, 
            current_price_set_back
        )
        return price_container

    @property
    def feature_container(self) -> FeaturesContainer:
        current_feature_set = (
            self.features
                .closest_head_omitted()
        )
        current_feature_set_back = (
            self.features
                .closest_peakback_by_omitted()
        )
        feature_container = FeaturesContainer(
            current_feature_set,
            current_feature_set_back
        )
        return feature_container

    @property
    def allocation_container(self):
        meta = self.technical.meta_model
        abbv = meta.abbreviation
        allocation = self.portfolio.current_allocation(
            symbol=abbv, is_mocked=self.is_mock)
        fill_pct = self.portfolio.current_pct_worth(
            symbol=abbv, is_mocked=self.is_mock)
        allocation_container = AllocationContainer(allocation, fill_pct)
        return allocation_container

    @property
    def is_next_features(self) -> bool:
        return self.features.is_next

    def change_stepsize(self, microseconds: float = 1000, seconds: float = 0, minutes: float = 0, hours: float = 10, days: float = 0, weeks: float = 0) -> None:
        """Changes the step size for all variables.

        Args:
            microseconds (float, optional): milliseconds. Defaults to 1000.
            seconds (float, optional): seconds. Defaults to 0.
            minutes (float, optional): minutes. Defaults to 0.
            hours (float, optional): hours. Defaults to 10.
            days (float, optional): days. Defaults to 0.
            weeks (float, optional): weeks. Defaults to 0.
        """
        local_time = self.time
        local_time.change_stepsize(microseconds=microseconds, seconds=seconds,
                                   minutes=minutes, hours=hours, days=days, weeks=weeks)

    def change_lookback(self, microseconds: float = 1000, seconds: float = 0, minutes: float = 0, hours: float = 10, days: float = 0, weeks: float = 0) -> None:
        """Changes the lookback time for the simulation.

        Args:
            microseconds (float, optional): milliseconds. Defaults to 1000.
            seconds (float, optional): seconds. Defaults to 0.
            minutes (float, optional): minutes. Defaults to 0.
            hours (float, optional): hours. Defaults to 10.
            days (float, optional): days. Defaults to 0.
            weeks (float, optional): weeks. Defaults to 0.
        """
        local_time = self.time
        local_time.change_lookback(microseconds=microseconds, seconds=seconds,
                                   minutes=minutes, hours=hours, days=days, weeks=weeks)

    def step(self):
        feature_container = self.feature_container
        price_container = self.price_container
        allocation_container = self.allocation_container
        ss = self.space.step(allocation_container,
                             feature_container, price_container)
        return ss

    def reset(self):
        self.technical.reset()



import datetime
from pathlib import Path
from typing import Callable, Optional, Union

import numpy as np
import pandas as pd
import pytest
import ta
from loguru import logger
from pandas.api.types import is_datetime64_any_dtype as is_datetime

from darwin_ml import __version__
from darwin_ml.technical import (fibonacci, fibonacci_rsi,
                                 super_hyper_mega_average_true_range)
from darwin_ml.technical.momentum import rsi_positions
from darwin_ml.technical.signals import fib_intensity_signal
from darwin_ml.technical.volume import fibonacci_boll_bands
from darwin_ml.utils import Windowing
from darwin_ml.utils.preprocessing import (format_look_ahead,
                                           format_timeseries_dataframe)




def format_all_ta(df, close="Close", stdev:int=2, window=20, dropset=[]):
    df = ta.utils.dropna(df)
    df.dropna()
    df = fibonacci(df, stdev=stdev, target=close, window=window)
    df = fibonacci_rsi(df, stdev=stdev, target=close, window=window)
    df = df.drop(columns=dropset)
    return df




from typing import Optional, Dict, Any
from loguru import logger

class PriceContainer(object):
    def __init__(self, current_bar: Dict[str, Any], previous_bar: Optional[Dict[str, Any]] = None):
        self.is_bar(current_bar)
        self._current = current_bar
        self._previous = previous_bar

        self._cprice = current_bar.get("close", 0.0)

        if bool(previous_bar):
            self.is_bar(previous_bar)
            self._previous = previous_bar
            self._pprice = previous_bar.get("close", 0.0)

    @property
    def current(self) -> Dict[str, Any]:
        """Get the latest bar.

        Returns:
            Dict[str, Any]: The latest bar.
        """
        return self._current

    @property
    def previous(self) -> Dict[str, Any]:
        """The previous price bar.

        Returns:
            Dict[str, Any]: The previous price bar.
        """
        return self._previous

    @property
    def is_none(self) -> bool:
        """Checks if both variables are empty.

        Returns:
            bool: Are both variables there.
        """
        is_current = bool(self.current)
        is_previous = bool(self.current)
        return bool(self.current) == False and bool(self.previous) == False

    @property
    def is_current(self) -> bool:
        """Determines if the event has only the current bar.

        Returns:
            bool: if this is the current bar.
        """
        return bool(self.current)
    
    @property
    def is_previous(self) -> bool:
        """Determines if the event has only the current bar.

        Returns:
            bool: if this is the current bar.
        """
        return bool(self.previous)

    @property
    def is_both(self) -> bool:
        """Determines if the event has only the current bar.

        Returns:
            bool: if this is the current bar.
        """
        return bool(self.current) and bool(self.previous)

    def verify(self, is_both: bool = False, is_prev: bool = False) -> None:
        """Run through a series of verifications.

        1. Check if at least one of the variables have been set.
        2. If we're checking for both, yet both variables aren't there, we should exit.
        3. We're checking if the latest price information is there.

        

        Args:
            is_both (bool, optional): Determines if we're checking for both variables. Defaults to False.
            is_prev (bool, optional): Determines if we're checking for the previous bar. Defaults to False.

        Raises:
            AttributeError: None of the variables have been set.
            AttributeError: One of the two expected values isn't there.
            AttributeError: The current price isn't there.
        """
        # If none of the variables have been set. We raise an attribute error.
        if self.is_none:
            raise AttributeError("None of the variables have been set.")

        # If we're looking for both variables, and one isn't set, we raise a value error
        if (is_both or is_prev) and not self.is_both:
            raise AttributeError("One of the two expected values isn't there.")

        # Check if the current bar is here
        if not self.is_current:
            raise AttributeError("The current price isn't there.")

    def is_bar(self, _bar: dict):
        """Check if the dictionary is a bar

        Args:
            _bar (dict): The dictionary that we're checking.
        """
        assert isinstance(_bar, dict)
        assert 'close' in _bar
        assert isinstance(_bar['close'], float)

    @property
    def current_price(self) -> float:
        """Return the latest close price.

        Returns:
            float: Close price.
        """
        self.verify()
        return self._cprice

    @property
    def previous_price(self) -> float:
        """Return the last close price.

        Returns:
            float: The last close price.
        """
        self.verify(is_prev=True)
        return self._pprice
    
    @property
    def direction(self):
        """Get the direction change from the price between two points.

        Returns:
            int: 1 represents the price went up. 0 represents the price went down.
        """
        if self.current_price > self.previous_price:
            return 1
        return 0
    
    @property
    def percentage(self) -> float:
        """Get the percentage change between the current price and the last price. 

        Returns:
            float: The percentage change between the two points.
        """
        self.verify(is_both=True)
        return (((self.current_price - self.previous_price) / self.current_price) * 100)
    

    def __str__(self):
        return f"<PriceContainer is_previous={self.is_previous} is_current={self.is_current} close={self.current_price} >"


from typing import Optional
import numpy as np
import pandas as pd
from pydantic import BaseModel


class MetadataModel(BaseModel):
    # Get the average percentage change of the allocation
    name: Optional[str] = None
    category: Optional[str] = None
    subcategories: Optional[dict] = None
    metatype: Optional[str] = None
    submetatype: Optional[str] = None
    abbreviation: Optional[str] = None



import collections
import copy
import random
from datetime import datetime
import typing
from typing import Optional, Union, Iterable, List
from uuid import UUID, uuid4
import numpy as np
import pandas as pd
from creme import (anomaly, cluster, ensemble, linear_model, metrics,
                   naive_bayes, optim, preprocessing, proba, stats)
from creme.base import Estimator
from creme.base.anomaly import AnomalyDetector
from creme.compose import Pipeline
from creme.metrics.base import Metric
from creme.naive_bayes.base import BaseNB
from creme.stats import Bivariate, Univariate
from loguru import logger
from pydantic import BaseModel, Field

from darwin_ml.models import PriceResponse
from darwin_ml.models.creme_ml.pipelines import (HardendedClassifierModel,
                                                 PctHardRegressorModel)
from darwin_ml.models.creme_ml.transforms import PercentageChange, PredictionHistory
from darwin_ml.middleware import containers
from jamboree.utils.context import timecontext

def default_normal():
    return random.normalvariate(0, 0.3)


def default_log():
    return np.log(random.normalvariate(0.3, 0.3))


def default_uniform():
    return random.uniform(-1, 1)


def default_binary():
    return random.randint(0, 2)


def predictive_classifier() -> Pipeline:
    return Pipeline(preprocessing.StandardScaler(),
                    linear_model.PAClassifier())


def anomoly_detection_pipeline() -> anomaly.HalfSpaceTrees:
    return anomaly.HalfSpaceTrees(n_trees=6, height=4, window_size=5)


class ModelResponse(BaseModel):
    pred_class: int = 0
    pred_pct: float = 0.0
    mean_imbal_up: float = 0.0
    mean_imbal_down: float = 0.0
    var_imbal_up: float = 0.0
    var_imbal_down: float = 0.0
    m1: int = 0
    m2: int = 0
    m3: int = 0
    anomaly_score: float = 0.0
    weighted_recall:float = 0.0
    weighted_f1:float = 0.0
    accuracy:float = 0.0

    mae_regression:float=0.0
    smape_regression:float=0.0
    mse_regression:float=0.0

    def space(self):
        arr = [
            int(self.pred_class),
            self.pred_pct,
            self.mean_imbal_up,
            self.mean_imbal_down,
            self.var_imbal_up,
            self.var_imbal_down,
            self.m1,
            self.m2,
            self.m3,
            self.anomaly_score,
            self.weighted_recall,
            self.weighted_f1,
            self.accuracy,
            self.mae_regression,
            self.smape_regression,
            self.mse_regression
        ]
        return np.array(arr) 


class OnlineModels(BaseModel):
    class_pred_history: PredictionHistory = PredictionHistory()
    reg_pred_history: PredictionHistory = PredictionHistory()
    hard_regressor: HardendedClassifierModel = PctHardRegressorModel()
    hard_classifier: HardendedClassifierModel = HardendedClassifierModel()

    prior_hard_regressor: HardendedClassifierModel = PctHardRegressorModel()
    prior_hard_classifier: HardendedClassifierModel = HardendedClassifierModel()

    # Get rolling information of the incorrect imbalances.
    # Incorrect imbalances can only be fitted when the imbalances are incorrect.
    imbalance_up_score_rolling_mean: stats.EWMean = stats.EWMean()
    imbalance_up_score_rolling_var: stats.EWVar = stats.EWVar()

    # Get rolling information of the incorrect imbalances.
    # Incorrect imbalances can only be fitted when the imbalances are incorrect.
    imbalance_down_score_rolling_mean: stats.EWMean = stats.EWMean()
    imbalance_down_score_rolling_var: stats.EWVar = stats.EWVar()

    # Get the most common classifications over time
    roll_mod_1: stats.RollingMode = stats.RollingMode(window_size=3)
    roll_mod_3: stats.RollingMode = stats.RollingMode(window_size=5)
    roll_mod_4: stats.RollingMode = stats.RollingMode(window_size=7)

    # All Stats from Creme's ClassificationReport
    # This will be a base model as well
    # It'll extract all of the variables for the classification
    report: Metric = metrics.ClassificationReport()
    anon: AnomalyDetector = anomoly_detection_pipeline()

    # get the anomaly_score to detect when something is extremely off
    mae_reg: metrics.MAE = metrics.MAE()
    smape_reg: metrics.SMAPE = metrics.SMAPE()
    mse_reg: metrics.MSE = metrics.MSE()

    class Config:
        arbitrary_types_allowed = True

    def set_lag(self, lag=3):
        self.class_pred_history: PredictionHistory = PredictionHistory(lag=lag)
        self.reg_pred_history: PredictionHistory = PredictionHistory(lag=lag)

    def update_rolling_mode(self, change_delta: PriceResponse):
        self.roll_mod_1.update(change_delta.direction)
        self.roll_mod_3.update(change_delta.direction)
        self.roll_mod_4.update(change_delta.direction)
        m_1 = self.roll_mod_1.get()
        m_3 = self.roll_mod_3.get()
        m_4 = self.roll_mod_4.get()
        return m_1, m_3, m_4

    def update_anomaly_detection(self, X: dict):
        """Update Anomaly Detection

        Args:
            X (dict): The features from the enviornment for the ML algorithm.

        Returns:
            float: The anomaly score.
        """
        self.anon.fit_one(X)
        return self.anon.score_one(X)

    def fit_estimators(self, prior: dict, cdelta: PriceResponse) -> None:
        
        self.hard_regressor.fit_one(prior, cdelta.percentage)
        self.hard_classifier.fit_one(prior, cdelta.direction)

    def predict_estimators(self, current: dict):
        """
            # Predict Estimators

            Also store the prediction in history. This will likely be duplicated in Jamboree later.

        """
        pct_prediction = self.hard_regressor.predict_one(current)
        dir_prediction = self.hard_classifier.predict_one(current)
        dir_prediction = int(dir_prediction)
        self.reg_pred_history.update(pct_prediction)
        self.class_pred_history.update(dir_prediction)
        return pct_prediction, dir_prediction

    def evaluate_estimators(self, cdelta: PriceResponse, prior: dict):
        c_pred = self.class_pred_history.get()
        r_pred = self.reg_pred_history.get()
        # Add regression error metrics later
        c_res = cdelta.direction
        r_res = cdelta.percentage
        if c_pred is not None:
            if c_res != c_pred:
                # logger.success(self.class_pred_history.window)
                self.evaluate_incorrect_estimation(prior)
            self.report.update(c_pred, c_res)

        if r_pred is not None:
            self.mae_reg.update(r_res, r_pred)
            self.mse_reg.update(r_res, r_pred)
            self.smape_reg.update(r_res, r_pred)

    def evaluate_incorrect_estimation(self, prior: dict):
        """ Gets imbalance information on the variables that weren't correctly estimated."""
        prior_probabilities = self.prior_hard_classifier.predict_proba_one(
            prior)
        _up_prob = prior_probabilities[True]
        _down_prob = prior_probabilities[False]
        self.imbalance_up_score_rolling_mean.update(_up_prob)
        self.imbalance_up_score_rolling_var.update(_up_prob)

        self.imbalance_down_score_rolling_mean.update(_down_prob)
        self.imbalance_down_score_rolling_var.update(_down_prob)

    def transfer_estimator(self):
        self.prior_hard_regressor = copy.copy(self.hard_regressor)
        self.prior_hard_classifier = copy.copy(self.hard_classifier)

    def fit_predict_eval(self, current: dict, prior: dict,
                         cdelta: PriceResponse):
        self.transfer_estimator()
        self.fit_estimators(prior, cdelta)
        _pct, _direction = self.predict_estimators(current)
        self.evaluate_estimators(cdelta, prior)
        return self.report, _pct, _direction

    def step(self, feature_container:containers.FeaturesContainer, change_delta: PriceResponse) -> ModelResponse:
        """Given the features and the current delta information get the following:
        
        1. predictions
        2. stats on those predictions
        3. Get the metrics on each Estimator

        Args:
            current_features (dict): [description]
            prior_features (dict): [description]
            change_delta (PriceResponse): [description]
        """

        curr = feature_container.current
        prev = feature_container.previous
        # anon_score = self.update_anomaly_detection(curr)
        
        report, _pct, _class = self.fit_predict_eval(curr,
                                                    prev,
                                                    change_delta)

        weighted_recall = report.weighted_recall.get()
        weighted_f1     = report.weighted_f1.get()
        accuracy        = report.accuracy.get()
        mae_reg = self.mae_reg.get()
        mse_reg = self.mse_reg.get()
        smape_reg = self.smape_reg.get()

        _im_up = self.imbalance_up_score_rolling_mean.get()
        _im_upv = self.imbalance_up_score_rolling_var.get()

        _im_down = self.imbalance_down_score_rolling_mean.get()
        _im_downv = self.imbalance_down_score_rolling_var.get()

        mod_1, mod_2, mod_3 = self.update_rolling_mode(change_delta)
        response = ModelResponse(
            m1=mod_1,
            m2=mod_2,
            m3=mod_3,
            # anomaly_score=anon_score,
            pred_class = int(_class),
            pred_pct = float(_pct),
            mean_imbal_up=_im_up,
            mean_imbal_down=_im_down,
            var_imbal_down=_im_downv,
            var_imbal_up=_im_upv,
            weighted_recall=weighted_recall,
            weighted_f1=weighted_f1,
            accuracy=accuracy,
            mae_regression=mae_reg,
            smape_regression=mse_reg,
            mse_regression=smape_reg
        )
        return response


""" 
    Step though accessed information (should have accessed by now). 
    Get all of the predictions, statistical information on those predictions and train.
"""


def main():
    _online_models = OnlineModels()
    logger.critical(_online_models)


if __name__ == "__main__":
    main()



import pprint
import time
import uuid
import warnings
from abc import ABC, abstractmethod, abstractproperty
from typing import Any, Dict, List, Optional, Union

import maya
from jamboree import BaseHandler, Jamboree, TimeHandler
from jamboree.handlers.default import DataHandler
from jamboree.utils.support.search import querying
from loguru import logger
from pydantic import BaseModel

from darwin_ml.technical import fiboacci_general, fibonacci
from darwin_ml.utils import Injectable

warnings.simplefilter(action='ignore', category=FutureWarning)


class StateTransformerAbstract(Injectable, ABC):
    """StateTransformer

    Arguments:
        ABC {Abstract} -- An abstract to be inheritted by future classes
    """
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.__state: Optional[Any] = None
        self.__handler: Optional[BaseHandler] = None
        # State Model
        self.__smodel: Optional[BaseModel] = None
        # Parameter Model
        self.__pmodel: Optional[BaseModel] = None
        # Step Model
        self.__stmodel: Optional[BaseModel] = None
        self.__time = TimeHandler()
        self.__episode: str = "live"
        self.__live: bool = False

    @property
    def time(self) -> TimeHandler:
        """The time handler used to get the current time.

        Returns:
            TimeHandler -- The time handler with the state loaded into it.
        """
        self.__time.processor = self.processor
        self.__time['episode'] = self.episode
        self.__time['live'] = self.live
        self.__time.reset()
        return self.__time

    @property
    def current_time(self) -> float:
        return self.time.head

    @property
    def episode(self) -> str:
        return self.__episode

    @episode.setter
    def episode(self, _episode: str):
        self.__episode = _episode

    @property
    def live(self) -> bool:
        return self.__live

    @live.setter
    def live(self, _live: bool):
        self.__live = _live

    @abstractproperty
    def state_model(self):
        raise NotImplementedError("State model not implemented.")

    @abstractproperty
    def param_model(self):
        raise NotImplementedError("Parameter model not implemented.")

    @abstractproperty
    def step_model(self):
        raise NotImplementedError("Step model not implemented.")

    @abstractproperty
    def state(self) -> Union[type]:
        raise NotImplementedError("State property not set")

    @abstractmethod
    def set_params(self, **kwargs):
        raise NotImplementedError("Parameters not set")

    @abstractmethod
    def reset(self):
        raise NotImplementedError("Reset function not implemented")

    @abstractmethod
    def step(self) -> BaseModel:
        """Take a single step necessary to create a state space.

        Raises:
            NotImplementedError: If the abstract method wasn't overwritten

        Returns:
            BaseModel -- A type constrained model to return all of the variables
        """
        raise NotImplementedError("Step function not implemented")
    
    def display_state(self):
        ts = maya.MayaDT(self.current_time)
        ep = self.episode
        lv = self.live
        state_message = f"\n\n The current time: {ts}\n\n Current Episode: {ep}\n\n Is Live: {lv}"
        logger.debug(state_message)


class TAPriceData(DataHandler):
    """
        # TAPriceData

        A way to store ta_features for the.
    """
    def __init__(self):
        super().__init__()
        self['category'] = "markets"
        self['submetatype'] = "technical"




import random
import time
import uuid
from pprint import pprint
from typing import Any, List, Optional

import maya
import pandas as pd
from addict import Dict
from jamboree import Jamboree
from jamboree.handlers.complex.engines import FileEngine
from jamboree.utils.support.search import querying
from loguru import logger

from darwin_ml.models import OnlineModels, StateSpaceModel

# logger.disable(__name__)


class IncrementalEngine(FileEngine):
    """ 
        Save and load incremental learning models dynamically by metadata and ID.

        Use to get rolling statistics for the 

        Uses context to make decisions and provide some clean up during the process :
        ```
            with strategy as strat:
                strat.step(dataframe)
        ```
    """
    def __init__(self, processor=None, **kwargs):
        super().__init__(processor=processor, **kwargs)
        self.entity = "incremental_models"
        self.current_strategy = None

    def init_specialized(self, **kwargs):
        super().init_specialized(**kwargs)

    def verify(self):
        """ Disabling verify. """
        pass

    def open_context(self):
        if not self.file_reset:
            self.reset()

    def close_context(self):
        pass

    def enterable(self):
        """ Return the object we want to enter into """
        return self.models

    def custom_post_load(self, item):
        self.current_strategy = item

    @property
    def models(self):
        if self.current_strategy is not None:
            return self.current_strategy
        raise AttributeError("No model set and found")

    def file_from_dict(self, item: Dict):
        reloaded = IncrementalEngine(processor=self.processor,
                                     name=item.name,
                                     category=item.category,
                                     subcategories=item.subcategories,
                                     submetatype=item.submetatype,
                                     abbreviation=item.abbreviation)
        return reloaded

    def reset(self):
        super().reset()
        logger.warning(self.metaid)
        if self.file_reset == True and self.current_strategy is None:
            if self.blobfile is not None:
                self.current_strategy = self.blobfile


class SampleStrategy(object):
    def step(self, df: pd.DataFrame):
        """ Preprocess the dataframe and make decisions using the step function. """
        return random.choice([0, 1, 2])


# @logger.catch
def main_strategy():
    jam_processor = Jamboree()
    strat = StateSpaceModel()
    incremental_engine = IncrementalEngine(processor=jam_processor,
                                           name="SAMPLE_NAMES",
                                           category="SAMCATSASSSS",
                                           subcategories={"sam": "ple"},
                                           submetatype="OTHER_SMAPLE__",
                                           abbreviation="SAP",
                                           blobfile=strat)
    incremental_engine.reset()
    while True:
        # We'd `pick` a strategy by `episode`. We'd first get an episode, get the strategy, then get the strategy, then pass in the data
        first_model = incremental_engine.first(name="SAMPLE_NAMES",
                                               category="SAMCATSASSSS",
                                               submetatype="OTHER_SMAPLE__")
        with first_model as model:
            logger.warning(model)


if __name__ == "__main__":
    main_strategy()



import collections
from creme.stats import Univariate
from creme.stats import pearson



class PercentageChange(Univariate):
    """Measures the serial correlation.

    This method computes the Pearson correlation between the current value and the value seen ``n``
    steps before.
    """

    def __init__(self, lag):
        self.window = collections.deque(maxlen=lag)
        self.pct_window = collections.deque(maxlen=lag)
        self.lag = lag

    @property
    def name(self):
        return f'pct_change_{self.lag}'


    def update(self, x):
        # The correlation can be update once enough elements have been seen
        if len(self.window) == self.lag:
            change, _ = self.calculate_change(self.window[0], x)
            self.pct_window.append(change)

        # Add x to the window
        self.window.append(x)
        return self

    def calculate_change(self, previous: float, current: float):
        change = (((current - previous) / current) * 100) 
        if change > 0: return change, 1
        return change, 0


    def get(self):
        pct_len = len(self.pct_window)
        if pct_len > 0: return self.pct_window[-1]
        return 0
    
    def get_class(self):
        if len(self.window) == self.lag:
            if self.window[-1] > self.window[0]: 
                return 1
            else:
                return 0
        return 99



import numpy as np
import pandas as pd
from loguru import logger
from typing import Optional
from pydantic import BaseModel
from darwin_ml.models.prediction import PredictionModel
from darwin_ml.models import AllocationStepSystem
from darwin_ml.models import OnlineModels
from darwin_ml.models import PriceDeltaModel
from darwin_ml.models.creme_ml.transforms import FeatureWindow
from darwin_ml.middleware import containers
from jamboree.utils.context import timecontext
class StateSpaceModel(object):
    """Blob Storage State Space Model

        This is the directly stored model for the state space. 
        The goal is to store all of the relavent variables and gradients.
    """
    def __init__(self, prediction_lag=3):
        # Get the change of price over the given lag period
        self.price_delta_calc: PriceDeltaModel = PriceDeltaModel()
        # Get the statistical change for the allocations over time.
        self.allo_step: AllocationStepSystem = AllocationStepSystem()
        # Predict and get a stacked version of the predictions over time.
        self.prediction_step: OnlineModels = OnlineModels()
        # Use this to locally store features and determine features relavent to the predictions.
        self.feature_window: FeatureWindow = FeatureWindow(lag=prediction_lag)
        
        self.prediction_step.set_lag(lag=prediction_lag)
        self.price_delta_calc.set_lag(lag=prediction_lag)

        self.current_prev_feature = None
    
    @property
    def is_previous_feature(self) -> bool:
        self.current_prev_feature = self.feature_window.get()
        return bool(self.current_prev_feature)

    @logger.catch(reraise=True)
    def step(self, all_c:containers.AllocationContainer,
             feature_container: containers.FeaturesContainer, price_container: containers.PriceContainer):
        
        curr = feature_container.current

        self.feature_window.update(curr)
        if not self.is_previous_feature:
            return None


        delta = self.price_delta_calc.step(price_container)
        pred_model = self.prediction_step.step(feature_container, delta)
        allo_model = self.allo_step.step(all_c.percentage, all_c.filling)
        joined_space = np.concatenate([pred_model.space(), allo_model.space()])
        return joined_space

    def step_complete(self, 
                        current_allocation: float,
                        current_fill_percentage: float, 
                        current_features: dict, 
                        prior_features: dict,
                        prior_price: dict,
                        current_price: dict):
        """Run the step with the prior features and price in the parameters. 
        
        Used to make real life predictions and changes.

        Args:
            current_allocation (float): The current allocation cap of the current asset.
            current_fill_percentage (float): How much of the allocation limit is filled.
            current_features (dict): The features at the current timestamp. Used to make prediction of future.
            current_price (dict): The current price of the asset.

        Returns:
            [type]: [description]
        """
        delta = self.price_delta_calc.step(current_price)
        return None
    
    def reset():
        """reset Reset function

        Reset core functions.
        """
        pass


import random
import numpy as np
import pandas as pd
from pydantic import BaseModel, Field
from datetime import datetime
from uuid import UUID, uuid4


def default_normal():
    return random.normalvariate(0, 0.3)


def default_log():
    return np.log(random.normalvariate(0.3, 0.3)) 


def default_uniform():
    return random.uniform(-1, 1)

def default_binary():
    return random.randint(0, 2)

class PredictionModel(BaseModel):
    log_pct_up_prediction: float = Field(default_factory=default_normal)
    percentage_up_prediction: float = Field(default_factory=default_normal)
    classification: int = Field(default_factory=default_binary)
    classification_proba: float = Field(default_factory=default_uniform)
    # the imbalance of incorrect predictions. Some datasets skew to one end or another
    imbalance_up: float = Field(default_factory=default_uniform)
    imbalance_down: float = Field(default_factory=default_uniform)

    # Get the most common classifications over time
    rolling_class_mode_1: int = Field(default_factory=default_binary)
    rolling_class_mode_3: int = Field(default_factory=default_binary)
    rolling_class_mode_4: int = Field(default_factory=default_binary)

    # All Stats from Creme's ClassificationReport
    # This will be a base model as well
    # It'll extract all of the variables for the classification
    py_classification_report: int = Field(default_factory=default_binary)
    joint_log_likelihood_2: float = Field(default_factory=default_uniform)
    joint_log_likelihood_3: float = Field(default_factory=default_uniform)
    anomaly_score: float = Field(default_factory=default_uniform)
    rolling_pct_change_mean: float = Field(default_factory=default_uniform)
    rolling_pct_change_var: float = Field(default_factory=default_uniform)
    rolling_pct_change_sem: float = Field(default_factory=default_uniform)
    rolling_pct_change_skew: float = Field(default_factory=default_uniform)
    # get the anomaly_score to detect when something is extremely off


    def prediction_array(self) -> np.ndarray:
        arr = [
            self.log_pct_up_prediction,
            self.percentage_up_prediction,
            self.classification,
            self.classification_proba,
            self.imbalance_up,
            self.imbalance_down,
            self.rolling_class_mode_1,
            self.rolling_class_mode_3,
            self.rolling_class_mode_4,
            self.joint_log_likelihood_2,
            self.joint_log_likelihood_3,
            self.anomaly_score,
            self.rolling_pct_change_mean,
            self.rolling_pct_change_var,
            self.rolling_pct_change_sem,
            self.rolling_pct_change_skew
        ]
        return np.array(arr)


from darwin_ml import DarwinStateSpace


class DarwinUnittestSetup(DarwinStateSpace):
    def __init__(self):
        super().__init__()
    

    def localized_test_setup(self):
        """
            # Setup Local Test
            ---

            Setups the test locally. All of the data is on Kevin's laptop. Data is already there with the appropiate data.
        """
        self.processor = Jamboree()
        self.asset = "80de790b2d024654b76de0e76a623bcf"
        self.time.head = setup_time()
        try:
            self.change_stepsize(days=1, hours=0, microseconds=0)
            self.change_lookback(days=20, microseconds=0)
            if self.price_count > 0 and self.feature_count == 0:
                self.technical.time.head = time.time()
                self.technical.reset()
                self.technical.time.head = setup_time()
        except Exception:
            pass


import uuid
from abc import ABC
from typing import Optional
from jamboree import Jamboree
from jamboree.handlers import TimeHandler

class Injectable(ABC):
    def __init__(self):
        self.__processor: Optional[Jamboree] = None
        self.__asset: Optional[str] = None
        self.__episode: str = uuid.uuid4().hex
        self.__live: bool = False

    @property
    def processor(self) -> Jamboree:
        if self.__processor is None:
            raise ValueError("The processor needs to be set")
        return self.__processor

    @processor.setter
    def processor(self, __processor: Jamboree):
        self.__processor = __processor


class InjectableTime(Injectable):
    def __init__(self):
        # super().__init__()
        self.__asset: Optional[str] = None
        self.__episode: str = uuid.uuid4().hex
        self.__live: bool = False
        self.__time: TimeHandler = TimeHandler()

    @property
    def time(self) -> TimeHandler:
        """The time handler used to get the current time.

        Returns:
            TimeHandler -- The time handler with the state loaded into it.
        """
        self.__time.processor = self.processor
        self.__time['episode'] = self.episode
        self.__time['live'] = self.live
        self.__time.reset()
        return self.__time

    @property
    def episode(self) -> str:
        return self.__episode

    @episode.setter
    def episode(self, _episode: str):
        self.__episode = _episode

    @property
    def live(self) -> bool:
        return self.__live

    @live.setter
    def live(self, _live: bool):
        self.__live = _live
    
    @property
    def asset(self) -> str:
        if self.__asset is None:
            raise ValueError("The asset hasn't been set yet.")
        return self.__asset

    @asset.setter
    def asset(self, _asset: str):
        self.__asset = _asset


from abc import ABC
from typing import List, Dict, Any
from addict import Dict as ADict
# from addict import Dict

class ProcedureAbstract(ABC):
    """ 
        Procedures ensure everything is consistent.
    """

    def verify(self):
        """ Ensures we have all of the required variables in place."""
        raise NotImplementedError("Verify function not implmented")

    def extract(self):
        """ The item that will be serialized. """
        raise NotImplementedError("Extract function not implemented")


class NamedModelMetric(ABC):
    def __init__(self, name:str):
        self.name = name
    

    def get_metric(self, y_pred, y_actual) -> Dict:
        raise NotImplementedError("You need to have a way to get a metric")


class NamedModelMetricSet:
    """ A single place to hold all of the model metrics (in a set)"""
    def __init__(self):
        self.metric_set:List[NamedModelMetric] = []
    

    def metrics(self, y_, y) -> Dict:
        if len(self.metric_set) == 0:
            return {}
        metric_listing = {}
        for metric in self.metric_set:
            name = metric.name
            metric_output = metric.get_metric(y_, y)
            metric_listing[name] = metric_output
        return metric_listing
    

    
class ProcedureManagement(ABC):
    """ A way to interact with procedures. Use to embed things like:
        
        1. Accessing procedures for a given code base.
        2. Check for certain attributes within the class
        3. Declare what's acceptable
    """
    def __init__(self):
        self.required_attributes:List[str] = []


    @property
    def allowed(self) -> List[str]:
        raise NotImplementedError("You need to set the allowed keys we'll take.")

    
    def check_allowed(self, key:str):
        if key not in self.allowed:
            raise ValueError(f"{key} has to be of the allowed keys ... ")


    def access(self, key:str) -> 'ProcedureAbstract':
        """ Access the procedure we need. Returns a procedure given the key we set it."""
        raise NotImplementedError("You need to create an access procedure")
    
    
    def isattr(self, parentinstance:Any):
        """ Checks to see if all of the attributes are in the parent class instance. """
        if len(self.required_attributes) > 0:
            for attr in self.required_attributes:
                if not hasattr(parentinstance, attr):
                    cls_name = parentinstance.__class__.__name__
                    msg = f"{cls_name} does not have the attribute {attr}"
                    raise AttributeError(msg)



class ModelProcedureAbstract(ProcedureAbstract):
    # _dict = None
    def __init__(self):
        self._mod = None
        self._opt = None
        self._crit = None
        
        self._model_dict = ADict()
        self._model_dict.model = None
        self._model_dict.optimizer = None
        self._model_dict.criteria = None

        self._model_typing = ADict()
        self._model_typing.model = None
        self._model_typing.optimizer = None
        self._model_typing.criteria = None

        self._model_requirements = ADict()
        self._model_requirements.model = True
        self._model_requirements.optimizer = False
        self._model_requirements.criteria = False
        
        self.changed = False
        self.named_metric_set = NamedModelMetricSet()


    @property
    def dictionary(self):
        """ A dictionary with all of the model information contained inside. """
        return self._model_dict
    
    @dictionary.setter
    def dictionary(self, _md:ADict):
        """ Load in raw model dict information """
        self._model_dict.update(_md)
        # self.verify()

    @property
    def requirements(self) -> ADict:
        """ Return a dictionary of requirements for the ... checking requirements"""
        return self._model_requirements
    
    @requirements.setter
    def requirements(self, _md:ADict):
        """ Load in raw model dict information """
        self._model_requirements.update(_md)
    
    @property
    def types(self) -> ADict:
        return self._model_typing

    @types.setter
    def types(self, _mt:ADict):
        self._model_typing.update(_mt)

    """
        Verification
    """

    def verify_model_typing(self):
        """Check that none of the model types are none """
        for k, v in self.requirements.items():
            if not isinstance(v, bool):
                raise ValueError(f"Model Requirement \'{k}\' must be a boolean value")
            if v == True:
                if self.types[k] is None:
                    raise ValueError(f"\'{k}\' Cannot be None in typing delarations")
                if self.dictionary[k] is None:
                    raise ValueError(f"\'{k}\' Cannot be None inside of the main model dictionary")
    
    def verify_model_dict(self):
        """ Verify that """
        for name, _type in self.types.items():
            if name is None or _type is None:
                continue
            current_item = self.dictionary[name]
            if not isinstance(current_item, _type) and not issubclass(current_item, _type):
                raise TypeError(f"{name} is not an instance of {_type}")

    def verify(self):
        self.verify_model_typing()
        self.verify_model_dict()
    
    def is_valid_data(self, _data) -> bool:
        """ Determines if the data we're about to use is valid"""
        raise NotImplementedError("Data validation not implemented yet")

    def split(self, X, y, **params):
        raise NotImplementedError

    def fit(self, X, y, **params):
        raise NotImplementedError

    def partial_fit(self, X, y, **params):
        raise NotImplementedError
    
    def predict(self, X, **params):
        raise NotImplementedError

    def predict_proba(self, X, **params):
        raise NotImplementedError

    def score(self, X, y, **params):
        raise NotImplementedError

    def get_params(self, **params):
        raise NotImplementedError

    def set_params(self, **params):
        raise NotImplementedError

    def extract(self):
        """ Get a dictionary to save the model. Should be called in close """
        return self.dictionary

    @property
    def metrics(self):
        """ Given the information we have, return a set of metrics"""
        metric_set = self.named_metric_set.metrics(0, 0)
        return metric_set
    
    

    

if __name__ == "__main__":
    model_types = ADict()
    model_vals = ADict()
    model_types.model = bool
    model_types.optimizer = str
    model_types.criteria = str

    model_vals.model = False
    model_vals.optimizer = "str"
    model_vals.criteria = "str"

    base_model_procedure = ModelProcedureAbstract()
    base_model_procedure.mtypes = model_types
    base_model_procedure.mdict = model_vals
    base_model_procedure.verify()
    
    print(base_model_procedure)


from typing import Optional, Dict, Any

from jamboree.handlers  import TimeHandler
from darwin_ml.handlers import StateTransformerAbstract
from darwin_ml.handlers import PriceStateTransformer
from darwin_ml.utils import Injectable




class OnlineStateSpace(Injectable):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__price: Optional[PriceStateTransformer] = None
        self.__allocation: Optional[StateTransformerAbstract] = None
        self.__predictions: Optional[PriceStateTransformer] = None

        self.episode: Optional[str] = None
        self.userid: Optional[str] = None
        self.assetid: Optional[str] = None
        self.live: bool = False

    @property
    def price(self) -> StateTransformerAbstract:
        return self.__price

    @property
    def predictions(self) -> StateTransformerAbstract:
        return self.__predictions

    @property
    def allocation(self) -> StateTransformerAbstract:
        return self.__allocation

    def check_null(self):
        if self.episode is None:
            raise AttributeError("An episode needs to be set")
        if self.userid is None:
            raise AttributeError("A userid needs to be set")
        if self.assetid is None:
            raise AttributeError("An asset id needs to be set")

    def set_userid(self, userid: str):
        if not isinstance(userid, str): raise TypeError("UserId is not a string type")
        self.userid = userid
        return self

    def set_episode(self, episode: str):
        if not isinstance(episode, str): raise TypeError("EpisodeId is not a string type")
        self.episode = episode
        return self

    def set_live(self, live: bool):
        if not isinstance(live, bool): raise TypeError("Live is not a bool type")
        self.live = live
        return self

    def set_assetid(self, assetid: str):
        if not isinstance(assetid, str): raise TypeError("AssetId is not a string type")
        self.assetid = assetid
        return self

    def set_params(self,
                   user_id: Optional[str] = None,
                   episode: Optional[str] = None,
                   live=False,
                   assetid: Optional[str] = None):
        if user_id is not None: self.set_userid(user_id)
        if episode is not None: self.set_episode(episode)
        if live is not None: self.set_live(live)
        if assetid is not None: self.set_assetid(assetid)

    def reset(self):
        self.price.reset()
        self.allocation.reset()
        self.predictions.reset()
    
    def initialization(self):
        pass


    def step(self) -> Dict[str, Any]:
        allocation_stats = self.allocation.step()


import collections
from creme.stats import Univariate
from creme.stats import pearson

class PredictionHistory(Univariate):
    """
        Use this history to get an accurate understanding of how the machine learning model has performed. 
        This is a duplication for saving the predictions in redis. Using the step function for training.
    """

    def __init__(self, lag=3):
        self.window = collections.deque(maxlen=lag)
        self.lag = lag

    @property
    def name(self):
        return f'prediction_history_{self.lag}'


    def update(self, x):
        # Add x to the window
        self.window.append(x)
        return self



    def get(self):
        """Get the prediction made at the lag.

        Returns:
            Any: The prediction from the ML algo. Use to 
        """
        pct_len = len(self.window)
        if pct_len == self.lag: return self.window[0]
        return None


from addict import Dict
from loguru import logger
from sklearn.base import BaseEstimator
from sklearn.datasets import make_friedman2
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import DotProduct, WhiteKernel
from jamboree.middleware.procedures import ModelProcedureAbstract



class SklearnProcedure(ModelProcedureAbstract):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self.requirements.model = True
        self.requirements.criterion = False
        self.requirements.optimizer = False
        self.types.model = BaseEstimator
        self.changed = False
    


    @property
    def model(self) -> BaseEstimator:
        self.verify()
        return self.dictionary.model
        
    def set_params(self, **params):
        self.changed = True
        self.model.set_params(**params)

    @logger.catch
    def get_params(self):
        return self.model.get_params()
        
    @logger.catch
    def predict(self, X, **kwargs):
        return self.model.predict(X, **kwargs)
    
    @logger.catch
    def predict_proba(self, X, **kwargs):
        prediction = self.model.predict_proba(X, **kwargs)
        return prediction

    @logger.catch
    def partial_fit(self, X, y, **kwargs):
        self.changed = True
        self.model.partial_fit(X, y, **kwargs)

    def fit(self, X, y, **kwargs):
        self.changed = True
        self.model.fit(X, y, **kwargs)

class CustomSklearnGaussianProcedure(SklearnProcedure):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        kernel = DotProduct() + WhiteKernel()
        gpr = GaussianProcessRegressor(kernel=kernel, random_state=0)
        self.dictionary.model = gpr


if __name__ == "__main__":
    general_procedure = CustomSklearnGaussianProcedure()
    X, y = make_friedman2(n_samples=500, noise=0, random_state=0)
    general_procedure.fit(X, y)
    print(general_procedure.predict(X[:2,:], return_std=True))
    # print(general_procedure.get_params())
    print(general_procedure.extract())




from .core import ProcedureAbstract, ProcedureManagement, ModelProcedureAbstract 
# from .models import CremeProcedure, SklearnProcedure, TFKerasProcedure, TorchProcedure
# from .management import ModelProcedureManagement


import numpy as np
import pandas as pd
import ta
from typing import Optional
from loguru import logger


def fibonacci_boll_bands(df: pd.DataFrame,
                         stdev: int = 2,
                         target: str = "Close",
                         window: int = 20) -> pd.DataFrame:
    bb_core = ta.volatility.BollingerBands(close=df[target],
                                           n=window,
                                           ndev=stdev)
    bb_1 = ta.volatility.BollingerBands(close=df[target],
                                        n=window,
                                        ndev=stdev * 0.236)
    bb_2 = ta.volatility.BollingerBands(close=df[target],
                                        n=window,
                                        ndev=stdev * 0.382)
    bb_3 = ta.volatility.BollingerBands(close=df[target],
                                        n=window,
                                        ndev=stdev * 0.5)
    bb_4 = ta.volatility.BollingerBands(close=df[target],
                                        n=window,
                                        ndev=stdev * 0.618)
    bb_5 = ta.volatility.BollingerBands(close=df[target],
                                        n=window,
                                        ndev=stdev * 0.764)
    lowered = target.lower()

    def intar(num: int):
        if num > 0:
            return f"h{num}_{lowered}"
        elif num == 0:
            return f"c1_{lowered}"
        else:
            return f"l{abs(num)}_{lowered}"

    df[intar(6)] = bb_core.bollinger_hband()
    df[intar(5)] = bb_5.bollinger_hband()
    df[intar(4)] = bb_4.bollinger_hband()
    df[intar(3)] = bb_3.bollinger_hband()
    df[intar(2)] = bb_2.bollinger_hband()
    df[intar(1)] = bb_1.bollinger_hband()
    df[intar(0)] = bb_core.bollinger_mavg()
    df[intar(-1)] = bb_1.bollinger_lband()
    df[intar(-2)] = bb_2.bollinger_lband()
    df[intar(-3)] = bb_3.bollinger_lband()
    df[intar(-4)] = bb_4.bollinger_lband()
    df[intar(-5)] = bb_5.bollinger_lband()
    df[intar(-6)] = bb_core.bollinger_lband()

    # ---------------------------------------------------------------------------------------
    # ---------------------------------------------------------------------------------------
    # ---------------------------------------------------------------------------------------
    # ---------------------------------------------------------------------------------------

    df = ta.utils.dropna(df)
    pos = f'pos_{lowered}'
    ab_center = f"above_below_{lowered}"
    df[pos] = 0
    df[ab_center] = 1 # Default above the center
    tar = df[target]

    def dtar(_num: int):
        return df[intar(_num)]

    df[ab_center].loc[:] = np.where(tar < dtar(0), 0, df[ab_center])
    
    df[pos].loc[:] = np.where(dtar(6) < tar, 7, df[pos])
    df[pos].loc[:] = np.where(np.logical_and(dtar(5) < tar,
                                      dtar(6) > tar), 6, df[pos])

    df[pos].loc[:] = np.where(np.logical_and(dtar(4) < tar,
                                      dtar(5) > tar), 5, df[pos])

    df[pos].loc[:] = np.where(np.logical_and(dtar(3) < tar,
                                      dtar(4) > tar), 4, df[pos])

    df[pos].loc[:] = np.where(np.logical_and(dtar(2) < tar,
                                      dtar(3) > tar), 3, df[pos])

    df[pos].loc[:] = np.where(np.logical_and(dtar(1) < tar,
                                      dtar(2) > tar), 2, df[pos])

    df[pos].loc[:] = np.where(np.logical_and(dtar(0) < tar,
                                      dtar(1) > tar), 1, df[pos])

    # ---------------------------------------------------------------------------------------
    # ---------------------------------------------------------------------------------------
    # ---------------------------------------------------------------------------------------
    # ---------------------------------------------------------------------------------------

    df[pos].loc[:] = np.where(np.logical_and(dtar(0) > tar,
                                      dtar(-1) < tar), 0, df[pos])

    df[pos].loc[:] = np.where(np.logical_and(dtar(-1) > tar,
                                      dtar(-2) < tar), -1, df[pos])

    df[pos].loc[:] = np.where(np.logical_and(dtar(-2) > tar,
                                      dtar(-3) < tar), -2, df[pos])

    df[pos].loc[:] = np.where(np.logical_and(dtar(-3) > tar,
                                      dtar(-4) < tar), -3, df[pos])

    df[pos].loc[:] = np.where(np.logical_and(dtar(-4) > tar,
                                      dtar(-5) < tar), -4, df[pos])

    df[pos].loc[:] = np.where(np.logical_and(dtar(-5) > tar,
                                      dtar(-6) < tar), -5, df[pos])
    df[pos].loc[:] = np.where(dtar(-6) > tar, -6, df[pos])

    # ---------------------------------------------------------------------------------------
    # ----------------------------- Drop The Bands ------------------------------------------
    # ---------------------------------------------------------------------------------------

    rlist = list(range(-6, 7))
    df_cols = [intar(x) for x in rlist]
    df = df.drop(columns=df_cols)

    return df



import uuid
import time

import maya
import pandas as pd
import pytest
import matplotlib.pyplot as plt
import typing
from stochastic.continuous import FractionalBrownianMotion
from loguru import logger
from jamboree import Jamboree

from darwin_ml.handlers import TAPredictionTransformer
from darwin_ml.params import PriceSetupResponse
import pandas_datareader.data as web

@pytest.fixture
def setup_price() -> PriceSetupResponse:
    data_microsoft = web.DataReader(
        "MSFT", "yahoo", start="2010/1/1", end="2020/6/10"
    ).round(2)
    data_apple = web.DataReader(
        "AAPL", "yahoo", start="2010/1/1", end="2020/6/10"
    ).round(2)
    return None



def test_download_succeeded(setup_price):
    assert setup_price is not None, "Setup Price Returned Nothing. It should have a `PriceSetupResponse` returned."

# @pytest.fixture
# def technical_pre() -> TAPredictionTransformer:
#     transformer = TAPredictionTransformer()
#     transformer.processor = Jamboree()
#     transformer.episode = uuid.uuid4().hex
#     transformer.live = False
#     return transformer


# @pytest.fixture
# def setup_time():
#     new_time = maya.now().subtract(years=5)
#     return new_time._epoch


# @pytest.fixture
# def technical(technical_pre: TAPredictionTransformer, setup_time: float):
    
#     technical_pre.asset = "80de790b2d024654b76de0e76a623bcf"
#     technical_pre.time.head = setup_time
#     technical_pre.time.change_stepsize(days=3, hours=0, microseconds=0)
#     technical_pre.time.change_lookback(days=20, microseconds=0)
#     technical_pre.reset_time()
#     return technical_pre


# def iterate_ta(ta_asset: TAPredictionTransformer):
#     pass


# def test_generate_stochastic(setup_price):
#     assert True, "Prices weren't generated"
#     assert True, "Prices weren't saved"


# def test_save_ta(setup_price, technical):
#     assert True, "TA Transformation Didn't Happen"
#     assert True, "Transformed TA features weren't saved"
#     assert True, "Test get recent price information"


# def test_load_technicals(technical: TAPredictionTransformer, setup_time):
#     """Test loading and iterating technical analysis.

#         1. Get time and manipulate it
#             - We should be able to set the time to now then check to see if the time matches.
#             - We should be able to set the lookback time by 
        
#         2. Store data to the system (APPL data)
#         3. That we're able to transform the data.
#         4. That we're able to retrieve the data from the head (it should have all of the expected fields)
#         5. That we're able to step forward and get the technical analysis features from a separate time period.
#             - The features should be different from the previous set of features.
        

#     Arguments:
#         setup_ta {TATransformer} -- Technical Analysis Tranformer
#         setup_time {float} -- Time epoch representing 5 years before.
#     """
#     abs_current = time.time()
#     price_count = technical.price.count()
    
#     if price_count > 0 and technical.feature_num == 0:
#         technical.time.head = abs_current
#         technical.reset()
#         technical.time.head = setup_time
    
#     assert (
#         abs_current - 100
#     ) > technical.current_time, "The current time should be much less than the current epoch time (in seconds)"

#     assert technical.price.count(
#     ) > 0, "There are no records inside of the database."

#     first_bar = technical.price.closest_head()
#     assert bool(
#         first_bar), "The closest bar is not none. This is actually very good."

#     assert "close" in first_bar

#     technical.time.step()
#     second_bar = technical.price.closest_head()

#     assert bool(
#         second_bar
#     ), "The second bar didn't have any information inside of the dictionary."
#     assert "close" in second_bar

#     first_close = first_bar.get("close")
#     second_close = second_bar.get("close")

#     assert first_close != second_close, "The bar information shouldn't be the exact same yet it is"


# def test_target_transition(technical: TAPredictionTransformer, setup_time: float):
#     """Get price delta information
    
#         1. load the price information.
#         2. create an iteration loop and continously get the current price and the previous price
#         3. Get the statistical information regarding each piece of the change.
#             - The percentage change between the prior price and current price.
#             - The change delta (is it going up or down?)
#                 - The changes should not be zero (that would indicate the price didn't change.)
#         4. Append the statistical information to a list
#             - Is the stat information growing over time (length of array)

#     """
#     technical.live = False
#     technical.time.head = setup_time
#     index = 0
#     while technical.features.is_next:
#         current_price_bar = technical.features.closest_head()
#         if not bool(current_price_bar):
#             break
#         print(current_price_bar)
#         technical.time.step()
#         index += 1
#         break
    
#     assert index > 0, "We didn't iterate over the technical analysis information more than once."


import pprint
import time
import uuid
import warnings
from typing import List

import maya
from jamboree import Jamboree
from jamboree.handlers.default import DataHandler
from jamboree.utils.support.search import querying
from loguru import logger

from darwin_ml.handlers import PriceStateTransformer
from darwin_ml.technical import fiboacci_general, fibonacci

warnings.simplefilter(action='ignore', category=FutureWarning)


class TAPriceData(DataHandler):
    """
        # TAPriceData

        A way to store ta_features for the.
    """
    def __init__(self):
        super().__init__()
        self['category'] = "markets"
        self['submetatype'] = "technical"

"""
def pick(self, _id: str):
self.clear()
current_search = self.search
item = current_search.pick(_id)
if item is None:
    raise AttributeError("File Metadata doesn't exist")

item = Dict(**item)
reloaded = self.file_from_dict(item)
reloaded.reset()
return reloaded
"""


class TAPredictionTransformer(PriceStateTransformer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__ta_price = TAPriceData()

    @property
    def features(self) -> TAPriceData:
        """TAPrice Data

        Raises:
            AttributeError: If there's not metadata loaded in just yet.

        Returns:
            DataHandler -- All of the processed TAInformation.
        """
        
        current_model = self.meta_model
        if current_model.name is None:
            raise AttributeError("No price metadata was set yet.")

        
        if self.is_same_meta_model():
            self.__ta_price.time = self.time
            self.__ta_price.episode = self.episode
            self.__ta_price.live = self.live
            return self.__ta_price
        # Uses the default price submetadata
        self.__ta_price.processor = self.processor
        self.__ta_price.time = self.time
        self.__ta_price.episode = self.episode
        self.__ta_price.live = self.live
        
        self.__ta_price['name'] = current_model.name
        self.__ta_price['category'] = current_model.category
        self.__ta_price['subcategories'] = current_model.subcategories
        self.__ta_price['metatype'] = current_model.metatype
        self.__ta_price['abbreviation'] = current_model.abbreviation
        self.__ta_price.reset()
        self.previous_meta_model = current_model
        return self.__ta_price
    
    @property
    def tcount(self) -> int:
        """Get a count of the number of TA Features.

        Returns:
            int -- The number of TA Features
        """
        return self.features.count()
    
    @property
    def feature_num(self) -> int:
        """Get a count of the number of TA Features.

        Returns:
            int -- The number of TA Features
        """
        return self.features.count()

    def preprocess_existing_price_info(self):
        curr_time = self.current_time
        self.time.head = time.time()
        df = self.price.dataframe_from_dynamic_peak(1, 100000)
        ta_features = fiboacci_general(df, target="close", high="high", _open="open", low="low", volume="volume", adj_close="adj_close")
        self.features.store_time_df(ta_features)
        self.time.head = curr_time

    def _reset_big_ta_preprocessing(self):
        """Preprocesses all price data
        """
        if not self.live:
            # check to see if we've already created all of the ta features for the given asset.
            logger.warning("This episode is not live")
            logger.success("We are running ta featurization.")
            logger.error(f"Number of features={self.feature_num}")
            if self.price.count() != 0:
                logger.critical("Creating technical analysis information")
                self.preprocess_existing_price_info()

    def _reset_price(self):
        """Resets the price information as a
        """
        if self.meta.name is None:
            # Load information here
            self.meta_model
            self.features
            
    def reset(self):
        self._reset_price()
        self._reset_big_ta_preprocessing()


if __name__ == "__main__":
    episode = uuid.uuid4().hex
    state_trans = TAPredictionTransformer()
    state_trans.asset = "80de790b2d024654b76de0e76a623bcf"
    state_trans.processor = Jamboree()
    state_trans.episode = episode
    state_trans.live = False
    # logger.success(state_trans.current_time)
    # logger.info(state_trans.meta_model)
    state_trans.reset()
    # logger.success(state_transformer.everything)
    # logger.warning(state_trans.step().prediction_array())



import numpy as np
import pandas as pd
import ta


def rsi_positions(df: pd.DataFrame,
                  target="Close",
                  window=14,
                  high=65,
                  low=36) -> pd.DataFrame:
    tag_decision_name = f"rsi_decision_{target.lower()}"
    tag_name = f"rsi_indicator_{target.lower()}"
    tag_shift = f"{tag_name}_shift"
    df = ta.utils.dropna(df)

    df[tag_name].loc[:] = ta.momentum.rsi(df[target])
    df[tag_shift].loc[:] = df[tag_name].shift(1)
    df = ta.utils.dropna(df)
    df[tag_decision_name] = 0
    df[tag_decision_name].loc[:] = np.where(np.logical_and(df[tag_name] > low, df[tag_shift] < low), 1, df[tag_decision_name])
    df[tag_decision_name].loc[:] = np.where(np.logical_and(df[tag_name] < high, df[tag_shift] > high), -1, df[tag_decision_name])
    
    df = df.drop(columns=[tag_shift, tag_name])
    return df



from .allocation import AllocationStepSystem
from .pricing import PriceResponse
from .metadata import MetadataModel
from .pricing import PriceDeltaModel 
from .online import OnlineModels
from .statespace import StateSpaceModel



# import warnings
# import pandas as pd
# import numpy as np
# from funhouse import TA
# import talib
# from sklearn import preprocessing
# from sklearn.model_selection import train_test_split
# # from emission.storages.pa_regressor.client import MemoryRegressorClient

# """ TA Preprocessing """
# # regress_client = MemoryRegressorClient(address="localhost", port=5581)
# def fib_feature(row):
    
#     # Using instead

#     main_value = row["price"]

#     # low7 = row["low7"]
#     low6 = row["low6"]
#     low5 = row["low5"]
#     low4 = row["low4"]
#     low3 = row["low3"]
#     low2 = row["low2"]
#     low1 = row["low1"]

#     up1 = row["up1"]
#     up2 = row["up2"]
#     up3 = row["up3"]
#     up4 = row["up4"]
#     up5 = row["up5"]
#     up6 = row["up6"]

#     current_value = 0

#     if low2 < main_value < low1:
#         current_value = -1
#     if low3 < main_value < low2:
#         current_value = -2
#     if low4 < main_value < low3:
#         current_value = -3
#     if low5 < main_value < low4:
#         current_value = -4
#     if low6 < main_value < low5:
#         current_value = -5
#     if main_value < low6:
#         current_value = -6

#     if up1 < main_value < up2:
#         current_value = 1
#     if up2 < main_value < up3:
#         current_value = 2
#     if up3 < main_value < up4:
#         current_value = 3
#     if up4 < main_value < up5:
#         current_value = 4
#     if up5 < main_value < up6:
#         current_value = 5
#     if main_value > up6:
#         current_value = 6

#     return current_value


# def rsi_feature(row, buy=35, sell=70):
#     buy_point = buy
#     sell_point = sell

#     if row < buy_point:
#         return 1
#     elif row > sell_point:
#         return -1
#     return 0




# """
#     Coin Price Preprocessing
# """



# def reduce_coin_tables(coin_data):
#     coin_data = coin_data.drop(["session_id", "type"], axis=1)
#     return coin_data


# def get_ta_features(coin_data):
#     ta = TA(coin_data).SMA(window=10).SMA(window=20).SMA(window=30).VolTrans(
#         window=20).RSI().ATR().FIBBB(stdev=1.7, window=20).EMA(window=10).EMA(window=20)
#     return ta


# def feature_engineer_fib(df, key_value="price"):
#     """ Find if the price or other value is in-between other values"""
#     df['fib_bet'] = df.apply(fib_feature, axis=1)
#     df = df.drop(labels=["up1", "up2", "up3", "up4", "up5", "up6",
#                          "basis", "low1", "low2", "low3", "low4", "low5", "low6"], axis=1)
#     return df


# def feature_engineer_main(df: pd.DataFrame):
#     # Get the rsi value
#     # Get the volatility RSI
#     df['rsi_buy_sell'] = df.RSI_14.apply(rsi_feature)
#     df['vol_buy_sell'] = df.voltrans_rsi.apply(rsi_feature, buy=40, sell=60)
#     return df


# def preprocess_moving_information(main_df: pd.DataFrame, fib_df: pd.DataFrame):
#     shifted_main = main_df.shift(-3)
#     shifted_price = shifted_main.price
#     fib_position = fib_df.fib_bet

#     main_df['fib_pos'] = fib_position
#     main_df['fib_trend'] = talib.ROCP(fib_position, timeperiod=2)
#     main_df['fib_trend_2'] = talib.ROCP(fib_position, timeperiod=4)
#     main_df['fib_trend_3'] = talib.ROCP(fib_position, timeperiod=6)

#     latest_3 = main_df.tail(3)

#     main_df['y'] = shifted_price
#     prediction_main = main_df.dropna()

#     prediction_main = prediction_main.drop(
#         columns=["price", "y"]).dropna()
#     weights = np.exp(np.linspace(-1., 0., len(prediction_main)))
#     weights_pred = np.exp(np.linspace(-1., 0., len(latest_3)))
#     latest_3["weights"] = weights_pred
#     prediction_main['weights'] = weights
#     prediction_main['y'] = shifted_price

#     return prediction_main, latest_3


# def scale_split(data_future):
#     prediction_main, latest_3 = data_future[0], data_future[1]
#     X = prediction_main.iloc[:, :-1].values
#     y = prediction_main.iloc[:, -1].values
#     X_pred = latest_3.iloc[:, :-1].values

#     X_scale = preprocessing.scale(X)
#     X_scale_pred = preprocessing.scale(X_pred)
#     return X_scale, X_scale_pred, y


# # def train_score_train_predict(scsplit_data, regress_select):
# #     X_scale, X_scale_pred, y = scsplit_data[0], scsplit_data[1], scsplit_data[2]
# #     score = 0
# #     prediction = []

# #     X_train, X_test, y_train, y_test = train_test_split(X_scale, y, test_size=0.2)
# #     regress_client.train(regress_select, X_train, y_train)
# #     score = regress_client.score(regress_select, X_test, y_test)
# #     regress_client.train(regress_select, X_test, y_test)
# #     prediction = regress_client.predict(regress_select, X_scale_pred)
# #     return score, prediction


# # with warnings.catch_warnings():
# #     warnings.simplefilter("ignore")



import math
import random
import typing
from datetime import datetime
from typing import Optional, Union
from uuid import UUID, uuid4

import numpy as np
import pandas as pd
from creme import (anomaly, cluster, compose, datasets, ensemble,
                   feature_extraction, feature_selection, linear_model, meta,
                   metrics, model_selection, naive_bayes, optim, preprocessing,
                   proba, sampling, stats)
from loguru import logger





class HardendedClassifierModel(object):
    def __init__(self):
        self.model =  (
            preprocessing.StandardScaler() |
            sampling.HardSamplingClassifier(
                classifier = linear_model.LogisticRegression(),
                 p = 0.15,
                 size = 300,
            )
        )


    def fit_one(self, x: dict, y: typing.Union[float, int]):
        self.model.fit_one(x, y)

    def predict_one(self, x: dict):
        return self.model.predict_one(x)
    
    def predict_proba_one(self, x:dict):
        return self.model.predict_proba_one(x)


import uuid
from typing import Optional

import pandas as pd
from jamboree import Jamboree, TimeHandler
from jamboree.handlers.abstracted.datasets import PriceData
from loguru import logger
from pydantic import BaseModel

from darwin_ml.handlers import StateTransformerAbstract
from darwin_ml.models.metadata import MetadataModel
from darwin_ml.models.prediction import PredictionModel


class PriceStateTransformer(StateTransformerAbstract):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__price = PriceData()
        self.__meta = MetadataModel()
        self.__assetid:Optional[str] = None
        self.previous_meta_model: Optional[MetadataModel] = None
        self.previous_asset: Optional[str] = None

    @property
    def price(self):
        if self.previous_asset == self.asset:
            return self.__price
        self.__price.processor = self.processor
        self.__price.episode = self.episode
        self.__price.live = self.live
        self.__price.pick(self.asset)
        self.previous_asset = self.asset
        return self.__price

    @property
    def asset(self) -> str:
        if self.__assetid is None or not isinstance(self.__assetid, str):
            raise ValueError("Assetid isn't set correctly")
        return self.__assetid

    @asset.setter
    def asset(self, _asset: str):
        self.__assetid = _asset

    @property
    def everything(self) -> pd.DataFrame:
        return self.price.everything

    @property
    def meta(self):
        return self.__meta

    @property
    def meta_model(self) -> MetadataModel:
        """Metadata Model

        Returns:
            MetadataModel -- The metadata model.
        """
        price_obj = self.price
        self.__meta.name = price_obj['name']
        self.__meta.category = price_obj['category']
        self.__meta.subcategories = price_obj['subcategories']
        self.__meta.metatype = price_obj['metatype']
        self.__meta.submetatype = price_obj['submetatype']
        self.__meta.abbreviation = price_obj['abbreviation']
        return self.__meta
    
    
    def is_same_meta_model(self):
        if self.previous_meta_model == self.meta_model:
            return True
        return False

    @property
    def is_next(self) -> bool:
        return self.price.is_next

    @property
    def state_model(self):
        """The BaseModel used to return a desired state. Use to provide type constraints.

        Returns:
            BaseModel -- The data we return back.
        """
        return None

    @property
    def param_model(self):
        """Supposed to be a pydantic model to state which parameters are to be there. Will delete soon.

        Returns:
            [type] -- [description]
        """
        return None

    @property
    def step_model(self):
        return None

    @property
    def state(self):
        return None

    def set_params(self,
                   asset_id: Optional[str] = None,
                   episode: Optional[str] = None,
                   live: Optional[bool] = None,
                   **kwargs):
        if asset_id is not None:
            self.asset = asset_id
        if episode is not None:
            self.episode = episode
        if live is not None:
            self.live = live

    def reset(self):
        # I don't know what's supposed to be here
        pass

    def step(self) -> PredictionModel:
        """Take a single step necessary to create a state space.

        Raises:
            NotImplementedError: If the abstract method wasn't overwritten

        Returns:
            BaseModel -- A type constrained model to return all of the variables
        """
        # logger.error(self.current_time)
        return PredictionModel()


if __name__ == "__main__":
    episode = uuid.uuid4().hex
    state_trans = PriceStateTransformer()
    state_trans.asset = "80de790b2d024654b76de0e76a623bcf"
    state_trans.processor = Jamboree()
    state_trans.episode = episode
    state_trans.live = False
    logger.success(state_trans.current_time)
    logger.info(state_trans.meta_model)
    # logger.success(state_transformer.everything)
    logger.warning(state_trans.step().prediction_array())



import numpy as np
import pandas as pd
import ta
from loguru import logger
from darwin_ml.technical import fibonacci_boll_bands
pd.set_option('mode.chained_assignment', None)


def ranged_exp_weight(arr):
    _arr = np.exp(np.abs(arr))
    return (_arr - np.min(_arr)) / np.ptp(_arr)


def fib_intensity_signal(df: pd.DataFrame, target="Close") -> pd.DataFrame:
    """Create a signal for the generated fiboacci position.

    Arguments:
        df {pd.DataFrame} -- OHLCV bar dataframe

    Keyword Arguments:
        target {str} -- What we're to be doing work on. Supposed to be generic and dynamic (default: {"Close"})

    Returns:
        pd.DataFrame -- A dataframe with the signals added in.
    """
    columns = list(df.columns)
    target = target.lower()
    tag = f"pos_{target}"
    tag_positive = f"pos_positive_{target}"
    shift_tag = f"{tag}_shifted"
    sig_tag = f"{tag}_signal"
    sig_ints = f"{tag}_intensity"
    sig_ints_two = f"{tag}_intensity_two"
    sig_ints_three = f"{tag}_intensity_three"
    sig_decimal = f"{tag}_decimal"
    _abc = f'above_below_{target}'
    importance = f'importance_{target}'

    if not tag in columns:
        logger.error(f"Missing given field: {tag}")
        return df

    df[shift_tag] = df[tag].shift(1)
    df = df.dropna()

    df[tag_positive] = 0
    df[tag_positive] = df[tag] + 7

    df[sig_tag] = 0
    df[sig_tag] = np.where((df[tag] == df[shift_tag]), 0, 0)
    df[sig_tag].loc[:] = np.where(df[tag] > df[shift_tag], 1, df[sig_tag])
    df[sig_tag].loc[:] = np.where(df[tag] < df[shift_tag], -1, df[sig_tag])

    roc_indicator = ta.momentum.ROCIndicator(close=df[tag_positive], n=5)
    roc_indicator_two = ta.momentum.ROCIndicator(close=df[tag_positive], n=8)
    roc_indicator_three = ta.momentum.ROCIndicator(close=df[tag_positive], n=3)

    df[sig_ints] = roc_indicator.roc() * 0.01
    df[sig_ints_two] = roc_indicator_two.roc() * 0.01
    df[sig_ints_three] = roc_indicator_three.roc() * 0.01
    
    df[sig_decimal] = df[tag] * 0.1
    tag_val = np.array(df[tag].values)
    df[importance] = ranged_exp_weight(tag_val)

    df = df.dropna()
    df = df.drop(columns=[shift_tag, tag_positive, tag])

    # Get signal here (buys and sells), drop the rest of the features here. Modify to allow for more features.
    return df


def fibonacci(df: pd.DataFrame,
              stdev: int = 2,
              target: str = "Close",
              window: int = 20):
    df = fibonacci_boll_bands(df, stdev=stdev, target=target, window=window)
    df = fib_intensity_signal(df, target=target)
    return df


def fibonacci_rsi(df: pd.DataFrame,
                  stdev: int = 2,
                  target: str = "Close",
                  window: int = 13):
    lts = target.lower()
    rsi_col = f"rsi_{lts}"
    df[rsi_col] = ta.momentum.RSIIndicator(df[target], n=window).rsi()
    df = fibonacci(df, target=rsi_col)
    df = df.drop(columns=[rsi_col])
    return df


def super_hyper_mega_average_true_range(df: pd.DataFrame,
                                        stdev: int = 2,
                                        target: str = "Close",
                                        high: str = "High",
                                        low: str = "Low",
                                        window: int = 20, rsi_window:int=14):
    lts = target.lower()
    tar_val = df[target]
    high_val = df[high]
    low_val = df[low]
    atr_col = f"atr_{lts}"
    df[atr_col] = ta.volatility.AverageTrueRange(
        high=high_val, low=low_val, close=tar_val, n=window).average_true_range()
    df = fibonacci_rsi(df, target=atr_col, window=rsi_window, stdev=stdev)
    df = df.drop(columns=[atr_col])

    return df


def fiboacci_general(df:pd.DataFrame,
                     stdev: int = 2,
                     atr_rsi_std:float=1.6,
                     target: str = "Close",
                     high: str = "High",
                     low: str = "Low",
                     _open:str="Open",
                     volume:str="Volume",
                     adj_close="Adj Close.",
                     window: int = 18,
                     atr_window:int=10,
                     rsi_atr_window: int = 7,
                     rsi_window: int = 7) -> pd.DataFrame:
    """Fibonacci General
    Does a massive number of transformations to the price data largely centered around the fibonacci bollinger bands. The imbalance
    Args:
        df (pd.DataFrame): OHLCV data indexed by time.
        stdev (int, optional): Standard Deviation. Defaults to 2.
        atr_rsi_std (float, optional): The ATR Standard Deviation. Defaults to 1.6.
        target (str, optional): The target number. Supposed to be the close value. Defaults to "Close".
        high (str, optional): The high value. Defaults to "High".
        low (str, optional): The low column name. Defaults to "Low".
        open (str, optional): The open column nam. Defaults to "Open".
        volume (str, optional): The volume column name. Defaults to "Volume".
        window (int, optional): The normal fibonacci window. Should be a relatively long horizon. Defaults to 45.
        atr_window (int, optional): The general ATR window. Defaults to 10.
        rsi_atr_window (int, optional): The size of the rsi for the ATR signal. Defaults to 14.
        rsi_window (int, optional): The size of the RSI window. Defaults to 14.
    """
    drop_cols = [target, high, low, _open, volume, adj_close]
    df_result = pd.DataFrame()
    _df_normal = df.copy(deep=True)
    _df_rsi = df.copy(deep=True)
    _df_atr = df.copy(deep=True)
    _df_normal = fibonacci(_df_normal, stdev=stdev, target=target, window=window)
    _df_rsi = fibonacci_rsi(_df_rsi, target=target, window=rsi_window)
    _df_atr = super_hyper_mega_average_true_range(_df_atr, stdev=atr_rsi_std, target=target, high=high, low=low, window=atr_window, rsi_window=rsi_atr_window)
    df_result = pd.concat([_df_normal,_df_rsi,_df_atr], axis=1)
    df_result = df_result.dropna()
    df_result = df_result.drop(columns=drop_cols)
    return df_result


import pandas as pd

def format_timeseries_dataframe(frame:pd.DataFrame, target:str) -> pd.DataFrame:
    """Sets a timeseries into the index of the dataframe.

    Arguments:
        frame {pd.DataFrame} -- The dataframe we're converting into a time series version (by index)
        target {str} -- The target variable we're to use as an index.

    Returns:
        pd.DataFrame -- A modified dataframe
    """
    columns = list(frame.columns)
    if target not in columns:
        raise AttributeError("The target variable isn't inside of the dataframe")
    timestamps = pd.to_datetime(frame[target], unit='s')
    frame.set_index(timestamps, inplace=True)
    frame = frame.drop(columns=[target])
    frame = frame.fillna(method="ffill")
    return frame

def format_look_ahead(frame:pd.DataFrame, target:str, size:int=1) -> pd.DataFrame:
    target_name = f"{target}_future"
    frame[target_name]  = frame[target].shift(size)
    frame.dropna()
    return frame


from collections import OrderedDict
from typing import Callable, Dict, List, Union

import creme
import numpy as np
import pandas as pd
from creme.base.estimator import Estimator
from creme.stream import iter_pandas
from loguru import logger


class Windowing:
    """Roll through a dataframe manually
    """
    def __init__(
        self,
        frame: pd.DataFrame,
        window_size: int = 10,
        adaptive_window: bool = False,
        adapted_window_size: int = 1,
    ):
        self._current_step: int = 0
        self._window_size: int = window_size
        self._lower_range: int = 0
        self._upper_range: int = len(frame)
        self._frame: pd.DataFrame = frame

        self._is_adapted = adaptive_window
        self._adapted_window_size = adapted_window_size

    @property
    def columns(self) -> list:
        """ Get Columns
            ---
            Return a list of columns from the dataframe.
        """
        return list(self.frame.columns)

    @property
    def window(self):
        """Get Window Size

        Returns:
            int -- The window size
        """
        if self.is_adaptive and self._current_step > self._window_size:
            return self.adapted_window
        return self._window_size

    @property
    def adapted_window(self) -> int:
        """Get the adaptive window size

        Returns:
            int -- Get window size
        """
        return self._adapted_window_size

    @property
    def is_adaptive(self) -> bool:
        return self._is_adapted

    @property
    def frame(self) -> pd.DataFrame:
        """Get the dataframe we're iterating through.

        Returns:
            pd.DataFrame -- The current dataframe
        """
        return self._frame

    @property
    def lower_bounds(self):
        return max((self._current_step - self.window, 0))

    @property
    def upper_bounds(self):
        return min(self._current_step + 1, len(self.frame))

    @property
    def has_next_observation(self) -> bool:
        """Returns if the dataframe has another observation

        Returns
        -------
        bool
            True if there is another observation.
        """
        return self._current_step < len(self.frame) - self.window - 1

    @property
    def next_observation(self) -> Union[np.ndarray, pd.DataFrame]:
        obs = self.frame[self.lower_bounds:self.upper_bounds]
        return obs

    def step(self):
        if self.has_next_observation:
            self._current_step += 1
            return self.next_observation
        raise IndexError("Why are you on the wrong index?")

    def reset(self):
        self._current_step = 0


class AdaptiveCremeTraining(Windowing):
    def __init__(
        self,
        model: Union[Estimator, OrderedDict],
        frame: pd.DataFrame,
        window_size: int = 10,
        adaptive_window: bool = False,
        adapted_window_size: int = 1,
    ):
        super().__init__(frame=frame,
                         window_size=window_size,
                         adaptive_window=adaptive_window,
                         adapted_window_size=adapted_window_size)
        """Adaptively train on a dataframe. 

        Will Stream through all elements in a dataframe if it's greater than 1. Otherwise, it'll train on the single item.
        The intention is to use this as a prototypical solution to training a single asset.

        Arguments:
            model {Union[Estimator, OrderedDict]} -- The model or pipeline we intend to use.
        """
        self.model: Union[Estimator, OrderedDict] = model
        self.metric_set = []

    def _is_valid_frame(self, frame: pd.DataFrame):
        if len(frame) > 0: return True
        return False
    
    def is_model(self):
        return self.model is not None

    def add_metric(self, metric):
        self.metric_set.append(metric)

    def step(self, frame: pd.DataFrame, target_column: Union[str, List[str]]):
        if not self._is_valid_frame(frame):
            return np.ndarray([[]])

        logger.debug(frame)



import unittest
from loguru import logger
import pandas_datareader.data as web

class TestFullStateSpace(unittest.TestCase):
    """This is a full backtest integration testing.

    Args:
        unittest: [description]
    """

    def setUp(self):
        logger.warning("download some test data from yfinance")
        logger.debug("Store that data and get the metadata id.")
        logger.success("Attach a userid")
        logger.error("Convert the data into technical analysis")
        logger.info("Convert the data into technical analysis")

    def tearDown(self):
        pass

    def test_upper(self):
        pass

    def test_isupper(self):
        pass

    def test_split(self):
        pass


if __name__ == '__main__':
    unittest.main()


import collections
from creme.stats import Univariate
from creme.stats import pearson





class FeatureWindow(Univariate):
    """
        # Feature Windowing Function

        Store all of the features in a window according to the number of steps we're predicting on.

        The idea here is to store all of the features relavent to the ml problem for a single call:
        ```py
            fch = FeatureChange(lag=3)
            feature_1 = {"feature_data": [1, 2, 3, 4, 5, 6]}
            feature_2 = {"feature_data": [4, 5, 6, 8, 9, 10]}
            feature_3 = {"feature_data": [1 ...]}
            feature_4 = {"feature_data": [2 ...]}
            fch.update(feature_1).update(feature_2).update(feature_3).update(feature_4)
            fch.get() 
            # Should return feature_2
            >>> {"feature_data": [4, 5, 6, 8, 9, 10]}
        ```
    """

    def __init__(self, lag):
        self.window = collections.deque(maxlen=lag)
        self.lag = lag

    @property
    def name(self):
        return f'feature_window_{self.lag}'


    def update(self, x):
        # Add x to the window
        self.window.append(x)
        return self

    def get(self):
        pct_len = len(self.window)
        if pct_len == self.lag or pct_len > 0: return self.window[0]
        return None



import uuid
import time

import maya
import pandas as pd
import pytest
import matplotlib.pyplot as plt
import typing
from stochastic.continuous import FractionalBrownianMotion
from loguru import logger
from jamboree import Jamboree

from darwin_ml.handlers import TAPredictionTransformer


class PriceSetupResponse:
    def __init__(self):
        self.__is_success: bool = False
        self.__metadata_id: typing.Optional[str] = None
        self.__microsoft_frame: Optional[pd.DataFrame] = None
        self.__apple_frame: Optional[pd.DataFrame] = None
    

    @property
    def success(self) -> bool:
        """success is_success property

        Were we able to successfully pull the price information.

        Returns:
            bool: Is_Success
        """
        is_success = self.__is_success and self.is_meta
        return is_success
    
    @success.setter
    def success(self, __success:bool):
        self.__is_success = __success

    @property
    def is_meta(self) -> bool:
        return self.metadata_id is not None


    @property
    def metadata_id(self):
        """metadata_id MetaData

        Metadata ID will be set if and only if the data was appropiately downloaded.

        Returns:
            str: The metadata id.
        """
        return self.__metadata_id
    
    @metadata_id.setter
    def metadata_id(self, _metadata_id:str):
        self.__metadata_id = _metadata_id
    

    @property
    def apple(self) -> pd.DataFrame:
        return self.__apple_frame
    

    @apple.setter
    def apple(self, value):
        self.__apple_frame = value

    @property
    def is_apple(self) -> bool:
        return self.apple is not None


    @property
    def microsoft(self):
        """The microsoft_dataframe property."""
        return self.__microsoft_frame

    @microsoft.setter
    def microsoft(self, value):
        self.__microsoft_frame = value

    @property
    def is_micro(self);
        return self.microsoft is not None




class DefaultPriceMetadataParameters:
    def __init__(self):
        self.category = "markets"
        self.subcategories = {
            "market": "stock",
            "country": "US",
            "sector": "techology",
        }
        self.submetatype = "price"